__copyright__ = """
Copyright (C)  2022 Rage Uday Kiran

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import time
import psutil
from tqdm import tqdm
import pandas as pd

class NumberImputation:
    def __init__(self, dataframe):
        self.df = dataframe.copy()
        self.df.columns = ['x', 'y'] + list(self.df.columns[2:])
        self.imputedDF = None

    def getStatistics(self, start_time):
        print("Total Execution time of proposed Algorithm:", time.time() - start_time)
        process = psutil.Process()
        memory_uss_kb = process.memory_full_info().uss / 1024
        print("Memory (USS) of proposed Algorithm in KB:", memory_uss_kb)
        memory_rss_kb = process.memory_full_info().rss / 1024
        print("Memory (RSS) of proposed Algorithm in KB:", memory_rss_kb)


    def impute(self, number = 0):
        start_time = time.time()
        xy = self.df[['x', 'y']].reset_index(drop=True)
        data = self.df.drop(['x', 'y'], axis=1).reset_index(drop=True)
        imputedData = data.fillna(number)
        self.imputedDF = pd.concat([xy, imputedData], axis=1)
        self.getStatistics(start_time)
        return self.imputedDF


    def save(self, outputFile='NumberImputation.csv'):
        if self.imputedDF is not None:
            try:
                self.imputedDF.to_csv(outputFile, index=False)
                print(f"Imputed data saved to: {outputFile}")
            except Exception as e:
                print(f"Failed to save labels: {e}")
        else:
            print("No imputed data to save. Run impute() first")


from geoAnalytics import csv2raster
from geoAnalytics.config import config
from shapely import wkb, wkt
from osgeo import gdal
import numpy as np
import pandas as pd
import psycopg2
import random
import os
import sys
from tqdm import tqdm
import multiprocessing
from scoreCalculator import ScoreCalculator

class GeoDatabaseManager:
    def __init__(self):
        self.conn = None
        self.scores = None

    def connect(self, dbName = None, hostIP = None, user = None, password = None, port=5432):
        """
        Connect to the database using the parameters.
        """
        if dbName != None:
            with open('database.ini', "w") as dbFile:
                dbFile.write("[postgresql]\nhost = " + str(hostIP) + "\nport = " + str(port) +
                             "\ndatabase = " + str(dbName) + "\nuser = " + str(user) + "\npassword = " + str(password))
        self.conn = self.test_connection()

    def disconnect(self):
        """
        Disconnect from the database.
        """
        if self.conn:
            print("Disconnecting from repository")
            self.conn.close()
            print("Disconnected from repository")

    def test_connection(self):
        """
        Test the connection to the database.
        """
        try:
            params = config()
            conn = psycopg2.connect(**params)
            curr = conn.cursor()
            curr.execute("select version();")
            for item in curr:
                print(item)
            print('You are now connected')
            return conn
        except (Exception, psycopg2.DatabaseError) as error:
            print(error)

    def create_repository(self, repositoryName, totalBands, SRID=4326):
        """
        Create a repository in the database.
        """
        try:
            params = config()
            self.conn = psycopg2.connect(**params)
            curr = self.conn.cursor()
            query = "geog geometry(POINT," + str(SRID) + "),"
            for i in range(1, totalBands):
                query += 'b' + str(i) + ' float,'
            query += 'b' + str(totalBands) + ' float'
            curr.execute(f"create table {repositoryName}({query})")
            curr.execute(f"CREATE INDEX index_{repositoryName} ON {repositoryName} USING GIST (geog)")
            self.conn.commit()
            print('Repository created')
        except (Exception, psycopg2.DatabaseError) as error:
            print(error)
        finally:
            self.disconnect()

    def delete_repository(self, repositoryName):
        """
        Delete a repository from the database.
        """
        try:
            params = config()
            self.conn = psycopg2.connect(**params)
            curr = self.conn.cursor()
            curr.execute(f"drop table {repositoryName};")
            self.conn.commit()
            print(f"{repositoryName} deleted successfully")
        except (Exception, psycopg2.DatabaseError) as error:
            print(error)
        finally:
            self.disconnect()
    
    def list_repositories(self):
        """
        List all repositories (tables) in the database.
        """
        try:
            self.connect()
            curr = self.conn.cursor()
            curr.execute(
                """
                SELECT table_name
                FROM information_schema.tables
                WHERE table_schema = 'public'
                """
            )
            repositories = curr.fetchall()
            print("Available Repos:\n---------------")
            for repository in repositories:
                print(repository[0])
            print("---------------")
        except (Exception, psycopg2.DatabaseError) as error:
            print(error)
        finally:
            self.disconnect()

    def clone_repository(self, repositoryName, cloneRepositoryName):
        """
        Clone a repository from the database.
        """
        try:
            params = config()
            self.conn = psycopg2.connect(**params)
            curr = self.conn.cursor()
            curr.execute(f"create table {cloneRepositoryName} as (select * from {repositoryName});")
            self.conn.commit()
            print('Repository cloned')
        except (Exception, psycopg2.DatabaseError) as error:
            print(error)
        finally:
            self.disconnect()

    def rename_repository(self, repositoryName, newRepositoryName):
        """
        Change the name of a repository in the database.
        """
        try:
            params = config()
            self.conn = psycopg2.connect(**params)
            curr = self.conn.cursor()
            curr.execute(f"ALTER TABLE IF EXISTS {repositoryName} RENAME TO {newRepositoryName};")
            self.conn.commit()
            print(f"{repositoryName} changed to {newRepositoryName}")
        except (Exception, psycopg2.DatabaseError) as error:
            print(error)
        finally:
            self.disconnect()
            
    def read_raster(self, fileName, totalBands, scalingFactor, SRID=4326):
        file = f"/tmp/{random.randint(0, 100000)}.txt"
        tempFile = self._r2tsv(totalBands, fileName, scalingFactor, SRID, file)
        df = pd.read_csv(tempFile, header = None, delimiter = " ")
        df.columns = ["WKB Position"] + ["B" + str(x) for x in range(totalBands)]
        return df
    
    def convertWKB(self, df):
        def wkb_to_geom(wkb_hex):
            return wkb.loads(bytes.fromhex(wkb_hex))

        # Apply the function to the 'wkb' column
        df['geometry'] = df['WKB Position'].apply(wkb_to_geom)

        # Extract longitude and latitude
        df['longitude'] = df['geometry'].apply(lambda geom: geom.x)
        df['latitude'] = df['geometry'].apply(lambda geom: geom.y)
        return df

    def insert_raster(self, repositoryName, fileName, totalBands, scalingFactor, SRID=4326):
        """
        Insert a TIFF file into the database.
        """
        tempFile = self._r2tsv(totalBands, fileName, scalingFactor, SRID)
        self.insert_csv(tempFile, repositoryName)

    def insert_raster_folder(self, repositoryName, folderName, totalBands, scalingFactor, extension=".lbl", SRID=4326, threads=16):
        """
        Insert TIFF files from a folder into the database.
        """
        files = [file for file in os.listdir(folderName) if file.endswith(extension)]
        print(files)
        
        pbar = tqdm(total=len(files))
        pool = multiprocessing.Pool(threads)
        
        def update(*a):
            pbar.update()
        
        for file in files:
            pool.apply_async(self.insert_raster, args=(repositoryName, os.path.join(folderName, file), totalBands, scalingFactor, SRID), callback=update)
        
        pool.close()
        pool.join()

    def insert_csv(self, filename, repositoryName, separator=' '):
        """
        Insert a CSV file into the database.
        """
        try:
            params = config()
            self.conn = psycopg2.connect(**params)
            curr = self.conn.cursor()
            # buffer = f"COPY {repositoryName} FROM '{filename}' DELIMITER '{separator}' CSV;"
            # buffer = f"\copy {repositoryName} FROM '{os.getcwd()}/{filename}' DELIMITER '{separator}' CSV;"
            with open(filename, 'r') as f:
                curr.copy_expert(f"COPY {repositoryName} FROM STDIN DELIMITER '{separator}' CSV", f)

            # curr.execute(buffer)
            self.conn.commit()
        except (Exception, psycopg2.DatabaseError) as error:
            print(error)
        finally:
            self.disconnect()

    def _r2tsv(self, endBand, srcfile, scalingFactor, SRID, tempFile = ""):
        """
        Convert a raster to a TSV file.
        """
        if tempFile == "":
            tempFile = f"/tmp/{random.randint(0, 100000)}.txt"
        band_nums = range(1, endBand + 1)

        srcds = gdal.Open(srcfile)
        if srcds is None:
            print(f'Could not open {srcfile}.')
            sys.exit(1)

        bands = [srcds.GetRasterBand(band_num) for band_num in band_nums]
        gt = srcds.GetGeoTransform()
        srcwin = (0, 0, srcds.RasterXSize, srcds.RasterYSize)

        with open(tempFile, 'wt') as dst_fh:
            band_format = ("%g " * len(bands)).rstrip() + '\n'

            for y in range(srcwin[1], srcwin[1] + srcwin[3]):
                data = [band.ReadAsArray(srcwin[0], y, srcwin[2], 1).reshape((srcwin[2],)) for band in bands]

                for x_i in range(0, srcwin[2]):
                    x = x_i + srcwin[0]
                    geo_x = gt[0] + (x + 0.5) * gt[1] + (y + 0.5) * gt[2]
                    geo_y = gt[3] + (x + 0.5) * gt[4] + (y + 0.5) * gt[5]
                    x_i_data = [data[i][x_i] * scalingFactor for i in range(len(bands))]
                    band_str = band_format % tuple(x_i_data)
                    line = f'{wkb.dumps(wkt.loads(f"POINT({round(geo_x, 4)} {round(geo_y, 4)})"), hex=True, srid=SRID)} {band_str}'
                    dst_fh.write(line)

        return tempFile

    def add_band(self, repositoryName, bandFormula):
        """
        Add a column to a table using the ALTER command.
        """
        try:
            params = config()
            self.conn = psycopg2.connect(**params)
            curr = self.conn.cursor()
            curr.execute(f"ALTER TABLE {repositoryName} ADD COLUMN B{bandFormula} float;")
            self.conn.commit()
            print('Band added to repository')
        except (Exception, psycopg2.DatabaseError) as error:
            print(error)
        finally:
            self.disconnect()

    def delete_band(self, repositoryName, bandNumber):
        """
        Delete the band number attribute from the table.
        """
        try:
            params = config()
            self.conn = psycopg2.connect(**params)
            curr = self.conn.cursor()
            curr.execute(f"ALTER TABLE {repositoryName} DROP COLUMN B{bandNumber};")
            self.conn.commit()
            print('Band deleted from repository')
        except (Exception, psycopg2.DatabaseError) as error:
            print(error)
        finally:
            self.disconnect()

    def get_raster(self, repositoryName, rasterFileName, Xmin, Ymin, Xmax, Ymax, Bands="*"):
        """
        Get a raster image from the database.
        """
        print('Getting raster from database')
        df = self.get_dataframe_for_envelope(repositoryName, Xmin, Ymin, Xmax, Ymax, Bands)
        self.dataframe_to_raster(df, rasterFileName)

    def dataframe_to_raster(self, dataframe, rasterFileName):
        """
        Create a raster image from a dataframe.
        """
        print('Creating raster file')
        csv2raster.csv2raster(output_file=rasterFileName, dataframe=dataframe)

    def get_dataframe_for_envelope(self, repositoryName, Xmin, Ymin, Xmax, Ymax, Bands="*"):
        """
        Get a dataframe for an envelope.
        """
        # ST_X(geog) as x, ST_Y(geog) as y
        sql = "SELECT ST_X(geog) as x, ST_Y(geog) as y, " + Bands + " FROM " + str(repositoryName) + " WHERE " + str(
            repositoryName) + ".geog && ST_MakeEnvelope(" + str(Xmin) + ',' + str(Ymin) + ',' + str(Xmax) + ',' + str(Ymax) + ");"
        params = config()
        self.conn = psycopg2.connect(**params)
        df = pd.read_sql(sql, self.conn)
        if Bands == "*":
            df.drop(columns=['geog'], inplace=True)
        return df
    
    def get_dataframe(self, repositoryName, Bands = "*"):
        """
        Get a dataframe for the entire repository.
        """
        sql = "SELECT ST_X(geog) as x, ST_Y(geog) as y, " + Bands + " FROM " + str(repositoryName) + ";"
        params = config()
        self.conn = psycopg2.connect(**params)
        
        df = pd.read_sql(sql, self.conn)
        if Bands == "*":
            df.drop(columns=['geog'], inplace=True)
        return df
    
    def filter(self, filterFile):
        """
        Filter the dataframe.
        """
        if self.scores is None:
            df = pd.read_csv(filterFile, header=None, delimiter=" ", dtype=float)
            self.scores = [ScoreCalculator(df[col].min(), df[col].max(), df[col].mean()) for col in df.columns]

    def calculate_scores_for_row(self, row):
    # Apply scoring only on relevant columns (from 3rd column onward)
        # range(2, len(row)) is used to skip the first two columns (x and y)
        # [j-2] is used to get the correct score object for the column because the first two columns are skipped aka x, y
        return [self.scores[j - 2].calculate_score(row[j]) for j in range(2, len(row))]


    def filtering(self, dataframe, filterFile):
        """
        Filter the dataframe and calculate scores for each row.
        """
        self.filter(filterFile)  # No need to store the result since it initializes self.scores

        # Apply score calculation to each row in the dataframe and expand results into new columns
        score_columns = dataframe.apply(self.calculate_scores_for_row, axis=1, result_type='expand')

        # Assign score columns to the dataframe
        dataframe[[f'score_{i}' for i in range(score_columns.shape[1])]] = score_columns

        return dataframe
    
    def total_score(self, dataframe):
        """
        Calculate the total score for the dataframe.
        """
        # add a new column to the dataframe that contains the sum of all score columns divided by the number of score columns
        num_score_cols = sum('score_' in col for col in dataframe.columns)
        
        dataframe['total_score'] = dataframe[[f'score_{i}' for i in range(1, num_score_cols)]].sum(axis=1) / num_score_cols

        return dataframe



import math
import os
import pandas as pd
import plotly.graph_objects as go
from plotly.subplots import make_subplots


class plotGUI:
    colorOptions = ['#a3a7e4', '#000000']
    sizeOptions = [10, 15]
    exportReady = pd.DataFrame()
    
    def drawEnvelope(dataframe):
        figuresEnvelope = go.FigureWidget(
            make_subplots(rows=2, cols=1, subplot_titles=("Click on point for band value information", "Values")))
        
        figuresEnvelope.update_layout(autosize=False,width=800,height=800)
        
        indexHeader = {"Values" : dataframe.columns[2:]}
        
        clicked = list()
        clickedNames = list()
        clickedDataframe = pd.DataFrame(indexHeader)
        clickedDataframe = clickedDataframe.set_index("Values")
        
        def updatePoint(trace,points,selector):
            pointIndex = points.point_inds[0]
            size = list(plotPoints.marker.size)
            pointData = dataframe.loc[pointIndex].tolist()
            point = str(pointData[0]) + " " + str(pointData[1])
            #if already clicked
            if point in clickedNames:
                clickedNames.remove(point)
                clickedDataframe.drop(columns=point, inplace=True)
                for i in points.point_inds:
                    size[i] = plotGUI.sizeOptions[0]
                    with figuresEnvelope.batch_update():
                        plotPoints.marker.size = size
            #if not clicked
            else:
                clickedDataframe[point] = pointData[2:]
                clickedNames.append(point)
                for i in points.point_inds:
                    size[i] = plotGUI.sizeOptions[1]
                    with figuresEnvelope.batch_update():
                        plotPoints.marker.size = size
                        
            #reset plot data
            figuresEnvelope.data = figuresEnvelope.data[:len(figuresEnvelope.data) - (len(figuresEnvelope.data) - 1)]
            
            #redraw plot
            for i in range(len(clickedDataframe.columns)):
                figuresEnvelope.add_trace(go.Scatter(x=clickedDataframe.loc[:, clickedDataframe.columns[i]].index,
                                                     y=clickedDataframe.loc[:, clickedDataframe.columns[i]],
                                                     mode='markers+lines', name=clickedNames[i], ), row=2, col=1)
            plotGUI.exportReady = clickedDataframe
        
        #add first plot to click
        figuresEnvelope.add_trace(
            go.Scatter(x=dataframe.x, y=dataframe.y, mode='markers', showlegend=False))
        
        
        plotPoints = figuresEnvelope.data[0]
        plotPoints.marker.color = [plotGUI.colorOptions[0]] * len(dataframe)
        plotPoints.marker.size = [plotGUI.sizeOptions[0]] * len(dataframe)
        plotPoints.on_click(updatePoint)
        display(figuresEnvelope)
    
    
    def drawNearestPoint(dataframe):
        figuresNearestPoint = go.FigureWidget(
            make_subplots(rows=2, cols=1, subplot_titles=("Click on point for band value information", "Values")))
        
        figuresNearestPoint.update_layout(autosize=False,width=800,height=800)
        
        indexHeader = {"Values" : dataframe.columns[2:]}
        
        clicked = list()
        clickedNames = list()
        clickedDataframe = pd.Dataframe(indexHeader)
        clickedDataframe = clickedDataframe.set_index("Values")
        
        def updatePoint(trace,points,selector):
            pointIndex = points.point_inds[0]
            size = list(plotPoints.marker.size)
            pointData = dataframe.loc[pointIndex].tolist()
            point = str(pointData[0]) + " " + str(pointData[1])
            #if already clicked
            if point in clickedNames:
                clickedNames.remove(point)
                clickedDataframe.drop(columns=point, inplace=True)
                for i in points.point_inds:
                    size[i] = plotGUI.sizeOptions[0]
                    with figuresNearestPoint.batch_update():
                        plotPoints.marker.size = size
            #if not clicked
            else:
                clickedDataframe[point] = pointData[2:]
                clickedNames.append(point)
                for i in points.point_inds:
                    size[i] = plotGUI.sizeOptions[1]
                    with figuresNearestPoint.batch_update():
                        plotPoints.marker.size = size
                        
            #reset plot data
            figuresNearestPoint.data = figuresNearestPoint.data[:len(figuresNearestPoint.data) -
                                                                (len(figuresNearestPoint.data) - 1)]
            
            #redraw plot
            for i in range(len(clickedDataframe.columns)):
                figuresNearestPoint.add_trace(go.Scatter(x=clickedDataframe.loc[:, clickedDataframe.columns[i]].index,
                                                         y=clickedDataframe.loc[:, clickedDataframe.columns[i]],
                                                         mode='markers+lines', name=clickedNames[i], ), row=2, col=1)
            plotGUI.exportReady = clickedDataframe
        
        #add first plot to click
        figuresNearestPoint.add_trace(
            go.Scatter(x=dataframe.x, y=dataframe.y, mode='markers', showlegend=False), row=1, col=1)
        
        #color the points
        distances = list()
        
        xPos = dataframe['x'][0]
        yPos = dataframe['y'][0]
        for i in range(len(dataframe)):
            distance = math.sqrt((abs(xPos - dataframe['x'][i]) *
                                  abs(xPos - dataframe['x'][i])) +
                                 (abs(yPos - dataframe['y'][i]) *
                                  abs(yPos - dataframe['y'][i])))
            distances.append(distance)
        
        def rgbhex(r, g, b):
            def clamp(x):
                return max(0, min(x, 255))
            return '#{0:02x}{1:02x}{2:02x}'.format(clamp(r), clamp(g), clamp(b))
        
        distanceColor = list()
        for i in range(len(dataframe)):
            value = int(distances[i] * max(300,300/len(dataframe)*300))
            distanceColor.append(rgbhex(255 - value, value, value))
        
        distanceColor[0] = plotGUI.colorOptions[1]
        sizeList = [plotGUI.sizeOptions[0]] * (len(dataframe))
        sizeList[0] = plotGUI.sizeOptions[1]

        plotPoints = figuresKNN.data[0]
        plotPoints.marker.color = tuple(distanceColor)
        plotPoints.marker.size = tuple(sizeList)
        
        figuresKNN.add_trace(go.Scatter(x=dataframe.iloc[0][2:].index, 
                                        y=dataframe.iloc[0][2:], mode='markers+lines',
                                        name=str(dataframe.iloc[0][0]) + " " + 
                                        str(dataframe.iloc[0][1]), ), row=2, col=1)
        
        plotPoints = figuresNearestPoint.data[0]
        
        clickedDataframe[str(dataframe.iloc[0][0]) + " " +
                  str(dataframe.iloc[0][1])] = dataframe.iloc[0]
        clickedNames.append(str(dataframe.iloc[0][0]) + " " + str(dataframe.iloc[0][1]))
        
        dataframe.drop([0],inplace=True)
        
        plotPoints.on_click(updatePoint)
        display(figuresNearestPoint)


try:
    from osgeo import gdal
except ImportError:
    import gdal

import sys

try:
    import numpy as Numeric
except ImportError:
    import Numeric


# =============================================================================
def Usage():
    print('Usage: raster2tsv.py [-skip factor] [-srcwin xoff yoff width height]')
    print('                   [-band b] srcfile [dstfile]')
    print('')
    sys.exit(1)


# =============================================================================
#
# Program mainline.
#
class raster2tsv:

    def __init__(self, parameter):
        # self.parameter = parameter
        srcwin = None
        skip = 1
        srcfile = None
        dstfile = None
        band_nums = []

        gdal.AllRegister()
        # argv = gdal.GeneralCmdLineProcessor(parameter)
        argv = parameter.split(" ")
        #print(argv)
        if argv is None:
            sys.exit(0)

        # Parse command line arguments.
        i = 0
        while i < len(argv):
            arg = argv[i]
            #print(arg)
            if arg == '-srcwin':
                srcwin = (int(argv[i + 1]), int(argv[i + 2]),
                          int(argv[i + 3]), int(argv[i + 4]))
                i = i + 4

            elif arg == '-skip':
                skip = int(argv[i + 1])
                i = i + 1

            elif arg == '-band':
                band_nums.append(int(argv[i + 1]))
                i = i + 1



            elif arg[0] == '-':
                Usage()



            elif srcfile is None:
                srcfile = arg


            elif dstfile is None:
                dstfile = arg

            else:
                Usage()

            i = i + 1

        if srcfile is None:
            Usage()
        if band_nums == []: band_nums = [1]
        # Open source file.
        srcds = gdal.Open(srcfile)
        if srcds is None:
            print('Could not open %s.' % srcfile)
            sys.exit(1)

        bands = []
        for band_num in band_nums:
            band = srcds.GetRasterBand(band_num)

            if band is None:
                print('Could not get band %d' % band_num)
                sys.exit(1)
            bands.append(band)

        gt = srcds.GetGeoTransform()

        # Collect information on all the source files.
        if srcwin is None:
            srcwin = (0, 0, srcds.RasterXSize, srcds.RasterYSize)

        # Open the output file.
        if dstfile is not None:
            dst_fh = open(dstfile, 'a')
        else:
            dst_fh = sys.stdout
        band_format = ("%g " * len(bands)).rstrip()

        # Setup an appropriate print format.
        if abs(gt[0]) < 180 and abs(gt[3]) < 180 \
                and abs(srcds.RasterXSize * gt[1]) < 180 \
                and abs(srcds.RasterYSize * gt[5]) < 180:
            format = '%.10g %.10g %s'

        else:
            format = '%.3f %.3f %s'

        # Loop emitting data.

        for y in range(srcwin[1], srcwin[1] + srcwin[3], skip):

            data = []
            for band in bands:
                band_data = band.ReadAsArray(srcwin[0], y, srcwin[2], 1)
                band_data = Numeric.reshape(band_data, (srcwin[2],))
                data.append(band_data)

            for x_i in range(0, srcwin[2], skip):

                x = x_i + srcwin[0]

                geo_x = gt[0] + (x + 0.5) * gt[1] + (y + 0.5) * gt[2]
                geo_y = gt[3] + (x + 0.5) * gt[4] + (y + 0.5) * gt[5]

                x_i_data = []
                for i in range(len(bands)):
                    x_i_data.append(data[i][x_i])

                band_str = band_format % tuple(x_i_data)

                line1 = format % (float(geo_x), float(geo_y), band_str)
                li = list(line1.split(" "))

                a = str(li[2:]).replace('\'', '').replace('[', '').replace(']', '')
                #print(a)
                temp_line = f"{li[0]}\t{li[1]}\t"#'Point(' +li[0] + ' ' + li[1] + ')\t'
                for a in li[2:]:
                    temp_line += str(a) + '\t'#temp_line = temp_line + str(a) + '\t'
                dst_fh.write(temp_line.rstrip('\t') + "\n")#dst_fh.write(temp_line[:-1] + "\n")


# Fuzzy C-Means (FCM) is a soft clustering algorithm where each data point can belong to multiple clusters with varying degrees of membership.
#
# **Importing this algorithm into a Python program**
#
#             from geoAnalytics.clustering import FuzzyCMeans
#
#             import pandas as pd
#
#             df = pd.read_csv("data.csv")
#
#             obj = FuzzyCMeans(df)
#
#             output = obj.clustering(n_clusters=3)
#
#             labels = output[0]
#
#             fcm.centers = output[1]
#
#             obj.save('FuzzyCMeansLabels.csv')
#


__copyright__ = """
Copyright (C)  2022 Rage Uday Kiran

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import time
import psutil
import numpy as np
import matplotlib.pyplot as plt
from tqdm import tqdm
from fcmeans import FCM
import pandas as pd


class FuzzyCMeans:
    """
    **About this algorithm**

    :**Description**: Fuzzy C-Means (FCM) is a soft clustering technique where each data point is assigned a membership score to all clusters,
                      allowing partial membership instead of a hard assignment to a single cluster.

    :**Parameters**: - **dataframe** (*pd.DataFrame*) -- * where:
        - The first two columns are coordinates.
        - The rest of the columns contain the features used for clustering.*

    :**Attributes**: - **df** (*pd.DataFrame*) -- *A copy of your original data, with the first two columns renamed to `'x'` and `'y'` to make things easy and clear during clustering*


    **Execution methods**

    **Calling from a Python program**

    .. code-block:: python

            from geoAnalytics.clustering import FuzzyCMeans

            import pandas as pd

            df = pd.read_csv("data.csv")

            obj = FuzzyCMeans(df)

            output = obj.clustering(n_clusters=3)

            labels = output[0]

            fcm.centers = output[1]

            obj.save('FuzzyCMeansLabels.csv')


     **Credits**

     The complete program was written by M. Charan Teja under the supervision of Professor Rage Uday Kiran.

     """

    def __init__(self, dataframe):
        """
        Constructor that initializes the FuzzyCMeans object with input DataFrame.

        :param dataframe:  - A Pandas DataFrame that contains the data to be clustered.
        """
        self.df = dataframe.copy()
        self.df.columns = ['x', 'y'] + list(self.df.columns[2:])
        self.labelsDF = None

    def getStatistics(self, start_time):
        """
        Prints memory usage and runtime of the clustering function.

        :param start_time: Time when the clustering function started execution.
        """
        print("Total Execution time of proposed Algorithm:", time.time() - start_time)
        process = psutil.Process()
        memory_uss_kb = process.memory_full_info().uss / 1024
        print("Memory (USS) of proposed Algorithm in KB:", memory_uss_kb)
        memory_rss_kb = process.memory_full_info().rss / 1024
        print("Memory (RSS) of proposed Algorithm in KB:", memory_rss_kb)

    def clustering(self, n_clusters=3):
        """
        Performs fuzzy clustering on data.
        -It excludes the 'x' and 'y' columns during clustering, and uses the remaining columns.
        -After clustering, it adds the cluster label (based on the highest membership score) back to the original coordinates
        :param n_clusters: Number of clusters to form (default is 3).
        :return: A tuple containing:
                 - A DataFrame with 'x', 'y', and cluster labels for each data point (based on maximum membership).
                 - A NumPy array with coordinates of the cluster centers.
        """
        start_time = time.time()
        data = self.df.drop(['x', 'y'], axis=1)
        data = data.to_numpy()
        fcm = FCM(n_clusters=n_clusters)
        fcm.fit(data)
        fcmResult = fcm.predict(data)
        label = self.df[['x', 'y']]
        self.labelsDF = label.assign(labels=fcmResult)
        self.getStatistics(start_time)
        return self.labelsDF, fcm.centers

    def save(self, outputFile='FuzzyCMeansLabels.csv'):
        """
        Save the outputFile in CSV
        """
        if self.labelsDF is not None:
            try:
                self.labelsDF.to_csv(outputFile, index=False)
                print(f"Labels saved to: {outputFile}")
            except Exception as e:
                print(f"Failed to save labels: {e}")
        else:
            print("No labels to save. Please run clustering first.")


import subprocess
import sys
from osgeo import gdal
import pandas as pd


class RasterConverter:
    """
    :Description: Convert raster data to CSV format and apply scale factor and offset to the converted data.

    Attributes:
        input_dataset (str): Path to the input raster dataset.
        output_csv (str): Path to the output CSV file.
        dataset (raster Dataset): GDAL dataset object representing the input raster dataset.
        raster_df (dataframe): DataFrame to store the converted data.


    """

    def __init__(self, input_dataset, output_csv):
        """
        Initializes the RasterConverter.

        Args:
            input_dataset (str): Path to the input raster dataset.
            output_csv (str): Path to the output CSV file.
        """
        self.input_dataset = input_dataset
        self.output_csv = output_csv
        self.dataset = None
        self.raster_df = None

    def getScalefactorOffset(self, band_number):
        """
        Retrieves the scale factor and offset for a given band.

        Args:
            band_number (int): The band number.

        Returns:
            tuple: The scale factor and offset for the band.
        """
        band = self.dataset.GetRasterBand(int(band_number))
        scale_factor = band.GetScale()
        offset = band.GetOffset()
        return scale_factor, offset

    def applyScaleOffset(self):
        """
        Applies the scale factor and offset to the converted data.
        """
        for i in range(len(self.raster_df.columns[2:])):
            scale_factor, offset = self.getScalefactorOffset(i + 1)
            self.raster_df[i + 2] = self.raster_df[i + 2] * scale_factor + offset

    def convert2csv(self):
        """
        Converts the raster dataset to CSV format.
        """
        command = ['gdal2xyz.py', '-all', self.input_dataset, self.output_csv]
        subprocess.call(command)

    def formatCoordinates(self):
        """
        Formats the first two columns of the DataFrame to "Point(x, y)" format.
        """
        self.raster_df[0] = "Point(" + self.raster_df[0].astype(str) + ","
        self.raster_df[1] = self.raster_df[1].astype(str) + ")"
        # self.raster_df[0] = self.raster_df[0].str.replace(r'\s', '')
        # self.raster_df[1] = self.raster_df[1].str.replace(r'\s', '')

    def runConversion(self):
        """
        Runs the entire conversion process.
        """
        self.convert2csv()

        self.raster_df = pd.read_csv(self.output_csv, header=None, sep=' ')

        self.dataset = gdal.Open(self.input_dataset)

        self.applyScaleOffset()

        self.formatCoordinates()

        print(self.raster_df.head())

        return self.raster_df


if __name__ == "__main__":
    input_dataset = sys.argv[1]
    output_csv = sys.argv[2]
    converter = RasterConverter(input_dataset, output_csv)
    rasterDf = converter.runConversion()
    rasterDf.to_csv(output_csv, sep=' ', index=None, header=None)


# KMeans++ enhances KMeans by improving initialization for better convergence and cluster quality, with added utilities for memory tracking, runtime reporting, and Elbow Method plotting on 2D-location DataFrame inputs.
#
# **Importing and Using this KMeans++ Wrapper in a Python Program**
#
#             import pandas as pd
#
#             from goeAnalytics.clustering import KMeansPP
#
#             df = pd.read_csv('data.csv')
#
#             obj = KMeansPP(df)
#
#             obj.elbowMethod()
#
#             output = obj.clustering(k=3)
#
#             labelsDF = output[0]
#
#             clusterCenters = output[1]
#
#             obj.save(outputFile='KMeansPPLabels.csv')
#

__copyright__ = """
Copyright (C)  2022 Rage Uday Kiran

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import time
import psutil
import numpy as np
import matplotlib.pyplot as plt
from tqdm import tqdm
from sklearn.cluster import KMeans as kmeansAlg
import pandas as pd


class KMeansPP:
    """
        **About this algorithm**

        :**Description**:KMeans++ improves K-Means clustering by using smarter centroid initialization for better stability and faster convergence, applied here to high-dimensional data excluding x, y coordinates.

        :**Parameters**:    - **dataframe** (*pd.DataFrame*) -- A Pandas DataFrame that contains the input dataset.
                            - The first two columns must be spatial or positional features (e.g., 'x' and 'y').
                            - All other columns are treated as feature vectors for clustering.

        :**Attributes**:    - **df** (*pd.DataFrame*) -- Stores the copy of the input dataset, renaming first two columns to 'x' and 'y'.
                            - **start_time** (*float*) -- Records the clustering start time for runtime analysis.
                            - **memory_uss_kb** (*float*) -- Measures USS memory usage in kilobytes after execution.
                            - **memory_rss_kb** (*float*) -- Measures RSS memory usage in kilobytes after execution.
                            - **labels** (*pd.DataFrame*) -- Final dataframe containing 'x', 'y', and cluster label for each instance.
                            - **cluster_centers_** (*np.ndarray*) -- Coordinates of the final cluster centroids after fitting.

        **Execution methods**

        **Calling from a Python program**

        .. code-block:: python

                import pandas as pd

                from goeAnalytics.clustering import KMeansPP

                df = pd.read_csv('data.csv')

                obj = KMeansPP(df)

                obj.elbowMethod()

                output = obj.clustering(k=3)

                labelsDF = output[0]

                clusterCenters = output[1]

                obj.save(outputFile='KMeansPPLabels.csv')

        **Credits**

        The complete program was written by Raashika and revised by M.Charan Teja under the supervision of Professor Rage Uday Kiran.

    """
    def __init__(self, dataframe):
        """
        Constructor to initialize the KMeans++ object with the given dataframe.
        """
        self.df = dataframe.copy()
        self.df.columns = ['x', 'y'] + list(self.df.columns[2:])
        self.labelsDF = None

    def getStatistics(self, start_time):
        """
        Prints memory usage and execution time after clustering.
        """
        print("Total Execution time of proposed Algorithm:", time.time() - start_time)
        process = psutil.Process()
        memory_uss_kb = process.memory_full_info().uss / 1024
        print("Memory (USS) of proposed Algorithm in KB:", memory_uss_kb)
        memory_rss_kb = process.memory_full_info().rss / 1024
        print("Memory (RSS) of proposed Algorithm in KB:", memory_rss_kb)

    def elbowMethod(self):
        """
        Applies the elbow method to help decide the optimal number of clusters (k).
        It plots WCSS (within-cluster sum of squares) for k in range 1 to 10.
        """
        wcss = []
        k_values = range(1, 11)
        data = self.df.drop(['x', 'y'], axis=1)

        for k in tqdm(k_values):
            kmeans = kmeansAlg(n_clusters=k, random_state=42, n_init=10)
            kmeans.fit(data)
            wcss.append(kmeans.inertia_)
        plt.plot(k_values, wcss, marker='o', linestyle='--')
        plt.xlabel('Number of Clusters (k)')
        plt.ylabel('WCSS')
        plt.title('Elbow Method for Optimal k (Ignoring Location Columns)')
        plt.show()

    def clustering(self, k = 4, max_iter=300):
        """
        Runs KMeans++ clustering on the input dataset using scikit-learn.

        :param k: Number of clusters to form.
        :param max_iter: Maximum number of iterations for a single run.
        :return: A DataFrame with original x, y and cluster labels, and the cluster centers.
        """
        start_time = time.time()
        data = self.df.drop(['x', 'y'], axis=1)
        data = data.to_numpy()
        kmeans = kmeansAlg(n_clusters=k, max_iter=max_iter, init='k-means++').fit(data)
        label = self.df[['x', 'y']]
        self.labelsDF = label.assign(labels=kmeans.labels_)
        self.getStatistics(start_time)
        return self.labelsDF, kmeans.cluster_centers_

    def save(self, outputFile='KMeansPPLabels.csv'):
        """
        Save the outputFile in CSV
        """
        if self.labelsDF is not None:
            try:
                self.labelsDF.to_csv(outputFile, index=False)
                print(f"Labels saved to: {outputFile}")
            except Exception as e:
                print(f"Failed to save labels: {e}")
        else:
            print("No labels to save. Please run clustering first.")


# GaussianMixture-based clustering algorithm using scikit-learn to assign cluster labels to multidimensional data with runtime and memory tracking, and support for saving results.
#
# **Importing and Using the Gaussianmixture Class in a Python Program**
#
#             import pandas as pd
#
#             from geoAnalytics.clustering import Gaussianmixture
#
#             df = pd.read_csv('input.csv')
#
#             gm = Gaussianmixture(df)
#
#             output = gm.clustering(n_components=3)
#
#             labels_df = output[0]
#
#             weights = output[1]
#
#             centers = output[2]
#
#             gm.save('GaussianMixtureLabels.csv')

__copyright__ = """
Copyright (C)  2022 Rage Uday Kiran

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import time
import psutil
import numpy as np
import matplotlib.pyplot as plt
from tqdm import tqdm
from sklearn.mixture import GaussianMixture
import pandas as pd


class Gaussianmixture:
    """
    **About this algorithm**

    :**Description**: Gaussian Mixture Model (GMM) is a probabilistic clustering algorithm that clusters feature-rich data by modeling it as a mixture of Gaussians, with runtime and memory tracking and exportable label results.

    :**Parameters**:    - Dataset (pandas DataFrame) must be provided during object initialization.
                        - No other parameters are required during instantiation.

    :**Attributes**:    - **df** (*pd.DataFrame*) -- The input data with 'x', 'y' coordinates and features.
                        - **labelsDF** (*pd.DataFrame*) -- DataFrame containing 'x', 'y', and assigned cluster labels.
                        - **model** (*GaussianMixture*) -- The trained scikit-learn GaussianMixture model instance for reuse or further analysis.


    **Execution methods**

    **Calling from a Python program**

    .. code-block:: python

            import pandas as pd

            from geoAnalytics.clustering import Gaussianmixture

            df = pd.read_csv("input.csv")

            gm = Gaussianmixture(df)

            output = gm.clustering(n_components=3)

            labels_df = output[0]

            weights = output[1]

            centers = output[2]

            gm.save('GaussianMixtureLabels.csv')

    **Credits**

    This implementation was created by Raashika and revised by M.Charan Teja under the guidance of Professor Rage Uday Kiran.
    """
    def __init__(self, dataframe):
        """
        Constructor to initialize the Gaussianmixture object with the input DataFrame.

        :param dataframe: pandas DataFrame containing at least columns ['x', 'y'] and feature columns.
        """
        self.df = dataframe.copy()
        self.df.columns = ['x', 'y'] + list(self.df.columns[2:])
        self.labelsDF = None

    def getStatistics(self, start_time):
        """
        Prints execution time and memory usage statistics of the clustering operation.

        :param start_time: Time.time() reference from before clustering.

        This method prints:
        - Total execution time in seconds.
        - Memory usage (USS - Unique Set Size) in kilobytes (KB), representing the memory
          uniquely used by the process.
        - Memory usage (RSS - Resident Set Size) in kilobytes (KB), representing the total
          physical memory used by the process.
        """
        print("Total Execution time of proposed Algorithm:", time.time() - start_time)
        process = psutil.Process()
        memory_uss_kb = process.memory_full_info().uss / 1024
        print("Memory (USS) of proposed Algorithm in KB:", memory_uss_kb)
        memory_rss_kb = process.memory_full_info().rss / 1024
        print("Memory (RSS) of proposed Algorithm in KB:", memory_rss_kb)

    def clustering(self, n_components=1, max_iters=100, covariance_type="full", init_params='kmeans', random_state=0):
        """
        Performs Gaussian Mixture Model clustering on the feature columns.

        :param n_components: Number of Gaussian components (clusters) to use.
        :param max_iters: Maximum number of iterations allowed during EM algorithm.
        :param covariance_type: Type of covariance parameters ('full', 'tied', 'diag', 'spherical').
        :param init_params: Initialization method ('kmeans' or 'random').
        :param random_state: Random seed to ensure reproducibility.
        :return: Tuple of (DataFrame with labels, array of component weights, array of component means).
        """
        start_time = time.time()
        data = self.df.drop(['x', 'y'], axis=1)
        data = data.to_numpy()
        gaussianMixture = GaussianMixture(n_components=n_components, max_iter=max_iters,
                                          covariance_type=covariance_type,
                                          init_params=init_params, random_state=random_state).fit(data)
        gaussianResult = gaussianMixture.predict(data)
        label = self.df[['x', 'y']]
        self.labelsDF = label.assign(labels=gaussianResult)
        self.getStatistics(start_time)
        return self.labelsDF, gaussianMixture.weights_, gaussianMixture.means_

    def save(self, outputFile='GaussianMixtureLabels.csv'):
        """
        Saves the clustering result (labels) into a CSV file.

        :param outputFile: Path to the output CSV file.
        """
        if self.labelsDF is not None:
            try:
                self.labelsDF.to_csv(outputFile, index=False)
                print(f"Labels saved to: {outputFile}")
            except Exception as e:
                print(f"Failed to save labels: {e}")
        else:
            print("No labels to save. Please run clustering first.")


import pandas as pd
import numpy as np
import time
import psutil
from tqdm import tqdm
from numba import njit, prange, cuda

class OneNNMaxNorm:
    def __init__(self):
        pass

    def getStatistics(self, start_time):
        print("Total Execution Time:", time.time() - start_time)
        process = psutil.Process()
        memory_kb = process.memory_full_info().uss / 1024
        print("Memory Usage (KB):", memory_kb)

    # ----------- Sequential Mode (Optimized) -----------
    def compute_maxnorm_sequential(self, testing, training):
        test_np = testing.to_numpy()
        train_np = training.to_numpy()

        if test_np.shape[1] != train_np.shape[1]:
            min_features = min(test_np.shape[1], train_np.shape[1])
            test_np = test_np[:, :min_features]
            train_np = train_np[:, :min_features]

        # Broadcasting: shape (num_test, 1, num_features) - (1, num_train, num_features)
        diff = np.abs(test_np[:, None, :] - train_np)
        max_diff = np.max(diff, axis=2)  # shape: (num_test, num_train)
        distances = np.min(max_diff, axis=1)
        return distances.tolist()

    # ----------- Parallel Mode -----------
    @staticmethod
    @njit(parallel=True)
    def compute_maxnorm_parallel(test_np, train_np):
        num_test = test_np.shape[0]
        num_train = train_np.shape[0]
        num_features = test_np.shape[1]
        distances = np.empty(num_test)

        for i in prange(num_test):
            min_dist = 1e10
            for j in range(num_train):
                max_diff = 0.0
                for k in range(num_features):
                    diff = abs(test_np[i, k] - train_np[j, k])
                    if diff > max_diff:
                        max_diff = diff
                if max_diff < min_dist:
                    min_dist = max_diff
            distances[i] = min_dist

        return distances

    # ----------- CUDA Mode -----------
    @staticmethod
    @cuda.jit
    def compute_maxnorm_cuda_kernel(testing, training, result):
        i = cuda.grid(1)
        if i >= testing.shape[0]:
            return

        num_features = testing.shape[1]
        min_dist = 1e10

        for j in range(training.shape[0]):
            max_diff = 0.0
            for k in range(num_features):
                diff = abs(testing[i][k] - training[j][k])
                if diff > max_diff:
                    max_diff = diff
            if max_diff < min_dist:
                min_dist = max_diff

        result[i] = min_dist

    def compute_maxnorm_cuda(self, testing, training):
        test_np = testing.to_numpy().astype(np.float32)
        train_np = training.to_numpy().astype(np.float32)

        if test_np.shape[1] != train_np.shape[1]:
            min_features = min(test_np.shape[1], train_np.shape[1])
            test_np = test_np[:, :min_features]
            train_np = train_np[:, :min_features]

        result = cuda.device_array(test_np.shape[0], dtype=np.float32)
        threads_per_block = 128
        blocks_per_grid = (test_np.shape[0] + threads_per_block - 1) // threads_per_block
        self.compute_maxnorm_cuda_kernel[blocks_per_grid, threads_per_block](test_np, train_np, result)

        return result.copy_to_host()

    # ----------- Entry Point -----------
    def run(self, training, testing, topK=-1, mode="sequential", algorithm="MaxNorm"):
        start_time = time.time()

        if algorithm == "difMaxNorm":
            training = training.diff(axis=1).iloc[:, 1:]
            testing = testing.diff(axis=1).iloc[:, 1:]

        if mode == "sequential":
            distances = self.compute_maxnorm_sequential(testing, training)
        elif mode == "parallel":
            distances = self.compute_maxnorm_parallel(testing.to_numpy(), training.to_numpy())
        elif mode == "cuda":
            if not cuda.is_available():
                raise RuntimeError("CUDA is not available on this machine.")
            distances = self.compute_maxnorm_cuda(testing, training)
        else:
            raise ValueError("Invalid mode. Choose 'sequential', 'parallel', or 'cuda'")

        testing = testing.copy()
        testing["1NNmaxNorm"] = distances
        sorted_df = testing.sort_values("1NNmaxNorm").head(topK)
        self.getStatistics(start_time)
        return testing, sorted_df


import os
import pandas as pd
from geoAnalytics.conversion import raster2tsv

class RasterToTSV:
    def __init__(self, input_file, output_file, start_band, end_band):
        self.input_file = input_file
        self.output_file = output_file
        self.start_band = start_band
        self.end_band = end_band
        #self.header = ['coordinate'] + [f'-band{band}' for band in range(start_band, end_band + 1)]
        self.header = ['x', 'y'] + [f'-band{band}' for band in range(start_band, end_band + 1)]

    def convert(self):
        if os.path.exists(self.output_file):
            os.remove(self.output_file)

        band_args = ' '.join([f'-band {b}' for b in range(self.start_band, self.end_band + 1)])
        params = f"{band_args} {self.input_file} {self.output_file}"

        print(f"Processing: {self.input_file}")
        raster2tsv.raster2tsv(params)  # class call with CLI-style string

        df = pd.read_csv(self.output_file, header=None, sep='\t')
        df.columns = self.header
        df.to_csv(self.output_file, index=False, sep='\t')

        print(f"Done. Output saved to: {self.output_file}")



import pandas as pd
import numpy as np


def cosineSimilarity(tableName1, tableName2):
    """
    :Description: This function calculates the cosine similarity between two dataframes using the provided formula.
    Cosine Similarity = Dot product of df1 and df2 / (Magnitude of df1 * Magnitude of df2)

    :param tableName1: The first data frame containing 'x' and 'y' columns (point co-ordinates)
    :param tableName2: The second data frame containing 'x' and 'y' columns (point co-ordinates)

    :return: A new DataFrame with 'x' , 'y' and 'cosine similarity' columns
    :rtype: DataFrame
    """

    df1 = tableName1
    df2 = tableName2

    cosineSimilarityDF = pd.DataFrame(columns=['X', 'Y', 'Cosine Similarity'])

    # get x and y
    x = df1['x']
    y = df1['y']

    # drop x and y
    df1 = df1.drop(['x', 'y'], axis=1)
    df2 = df2.drop(['x', 'y'], axis=1)

    # convert to numpy array
    df1 = df1.to_numpy()
    df2 = df2.to_numpy()

    # calculate cosine similarity

    # for each row in df2
    for i in range(len(df2)):
        similarity_scores = df1.dot(df2) / (np.linalg.norm(df1, axis=1) * np.linalg.norm(df2))
        # get average of similarity scores
        average_similarity_score = np.average(similarity_scores)
        # append to dataframe
        cosineSimilarityDF = cosineSimilarityDF.append(
            {'X': x[i], 'Y': y[i], 'Cosine Similarity': average_similarity_score}, ignore_index=True)

    return cosineSimilarityDF




from osgeo import gdal
import pandas as pd
import os
import random
import subprocess

def csv2raster(input_file='', output_file='output.nc', sep=" ", dataframe=''):
    input_file = input_file
    inputfile_sep = sep
    tempOut = output_file
    output_file = 'output.nc'
    dataFrame = dataframe
    if input_file != '':
        dataFrame = pd.read_csv(input_file, 
                                     sep=inputfile_sep, 
                                     header=None, index_col=None)


    dataFrame = dataFrame.sort_values(['y','x'], ascending=[True,True])
    dataFrameColumns = dataFrame.columns

    randInt = str(random.randint(0, 100000))
    dataFrame = dataFrame.sort_values(by=['y', 'x'], ascending = [False,True])

    for i in range(2,len(dataFrameColumns)):
        dataFrame.to_csv("xyzformat.xyz", index=False, header=None, sep=" ")
        raster = gdal.Translate("temp_" + randInt + "_" + str(dataFrameColumns[i]) + ".nc", "xyzformat.xyz")
        dataFrame = dataFrame.drop(columns=dataFrameColumns[i])
        buffer = 'ncrename -v ' + 'Band1,' + str(dataFrameColumns[i]) + " temp_" + randInt + "_" + str(dataFrameColumns[i]) + ".nc" 
        print(subprocess.getstatusoutput(buffer))


    buffer = 'cdo -f nc2 cat ' + "temp_" + randInt + "_*.nc " + output_file
    print(subprocess.getstatusoutput(buffer))

    buffer = 'rm ' + "temp_" + randInt + "_*.nc"
    print(subprocess.getstatusoutput(buffer))
    
    if tempOut[-3:] == '.nc':
        #rename
        os.rename(output_file, tempOut)
    elif tempOut[-3:] == 'iff' or tempOut[-3:] == 'tif':
        buffer = 'gdal_translate -of GTiff output.nc ' + tempOut
        print(subprocess.getstatusoutput(buffer))
        buffer = 'rm output.nc'
        print(subprocess.getstatusoutput(buffer))
    else:
        os.rename(output_file, tempOut + ".nc")


    os.remove("xyzformat.xyz")



import pandas as pd
import numpy as np
import time
import os
import psutil
from tqdm import tqdm
from numba import njit, prange, cuda

# =================== Numba-compatible Functions OUTSIDE the class ===================

@njit
def dtw_numba(A, B):
    N, M = len(A), len(B)
    d = np.zeros((N, M))
    for n in range(N):
        for m in range(M):
            d[n][m] = (A[n] - B[m]) ** 2
    D = np.zeros((N, M))
    D[0][0] = d[0][0]
    for n in range(1, N):
        D[n][0] = d[n][0] + D[n - 1][0]
    for m in range(1, M):
        D[0][m] = d[0][m] + D[0][m - 1]
    for n in range(1, N):
        for m in range(1, M):
            D[n][m] = d[n][m] + min(D[n - 1][m], D[n - 1][m - 1], D[n][m - 1])
    return D[N - 1][M - 1]

@njit(parallel=True)
def compute_dtw_parallel(testing_data, training_data):
    num_test = testing_data.shape[0]
    num_train = training_data.shape[0]
    distances = np.empty(num_test)
    for i in prange(num_test):
        min_dist = 1e10
        for j in range(num_train):
            dist = dtw_numba(testing_data[i], training_data[j])
            if dist < min_dist:
                min_dist = dist
        distances[i] = min_dist
    return distances


@njit
def dtw_manual(A, B):
    """Manually computes DTW distance using explicit formula."""
    N, M = len(A), len(B)
    d = np.zeros((N, M))
    for n in range(N):
        for m in range(M):
            d[n, m] = (A[n] - B[m]) ** 2
    D = np.zeros((N, M))
    D[0, 0] = d[0, 0]
    for n in range(1, N):
        D[n, 0] = d[n, 0] + D[n - 1, 0]
    for m in range(1, M):
        D[0, m] = d[0, m] + D[0, m - 1]
    for n in range(1, N):
        for m in range(1, M):
            D[n, m] = d[n, m] + min(D[n - 1, m], D[n - 1, m - 1], D[n, m - 1])
    return D[N - 1, M - 1]

@njit(parallel=True)
def compute_dtw_sequential_numba(testing_np, training_np):
    """Optimized sequential-threaded DTW computation using explicit formula."""
    num_test = testing_np.shape[0]
    num_train = training_np.shape[0]
    distances = np.full(num_test, np.inf)  # Preallocate with large values

    for i in prange(num_test):  # Parallel loop with Numba
        for j in range(num_train):
            dist = dtw_manual(testing_np[i], training_np[j])  # Using optimized DTW
            if dist < distances[i]:
                distances[i] = dist

    return distances

# =================== Main Class ===================

class OneNNDTW:
    @staticmethod
    def get_statistics(start_time):
        print("Total Execution Time:", time.time() - start_time)
        process = psutil.Process(os.getpid())
        memory_kb = process.memory_full_info().uss / 1024
        print("Memory Usage (KB):", memory_kb)

    @staticmethod
    @cuda.jit
    def dtw_cuda_kernel(testing_data, training_data, result):
        i = cuda.grid(1)
        if i < testing_data.shape[0]:
            min_dist = 1e10
            for j in range(training_data.shape[0]):
                dist = 0.0
                for k in range(testing_data.shape[1]):
                    val = testing_data[i][k] - training_data[j][k]
                    dist += val * val
                if dist < min_dist:
                    min_dist = dist
            result[i] = min_dist

    @staticmethod
    def compute_dtw_cuda(testing, training):
        test_np = testing.to_numpy().astype(np.float32)
        train_np = training.to_numpy().astype(np.float32)
        result = cuda.device_array(test_np.shape[0], dtype=np.float32)
        threads_per_block = 128
        blocks_per_grid = (test_np.shape[0] + threads_per_block - 1) // threads_per_block
        OneNNDTW.dtw_cuda_kernel[blocks_per_grid, threads_per_block](test_np, train_np, result)
        return result.copy_to_host()

    def compute_dtw_sequential(self, testing, training):
        """Wrapper to convert DataFrame to NumPy and apply optimized function."""
        testing_np = testing.to_numpy()
        training_np = training.to_numpy()
        distances = compute_dtw_sequential_numba(testing_np, training_np)
        return distances.tolist()

    def run(self, training, testing, topK=-1, mode="sequential", algorithm="DTW"):
        start_time = time.time()

        if algorithm == "difDTW":
            training = training.diff(axis=1).iloc[:, 1:]
            testing = testing.diff(axis=1).iloc[:, 1:]

        if mode == "sequential":
            distances = self.compute_dtw_sequential(testing, training)
        elif mode == "parallel":
            distances = compute_dtw_parallel(testing.to_numpy(), training.to_numpy())
        elif mode == "cuda":
            if not cuda.is_available():
                raise RuntimeError("CUDA is not available on this machine.")
            distances = self.compute_dtw_cuda(testing, training)
        else:
            raise ValueError("Invalid mode. Choose 'sequential', 'parallel', or 'cuda'")

        testing['1NNDTW'] = distances
        sorted_df = testing.sort_values('1NNDTW').head(topK)
        self.get_statistics(start_time)
        return testing, sorted_df



import numpy as np
import time
import psutil
from tqdm import tqdm
from sklearn.svm import OneClassSVM


class OneNNSVM:
    def __init__(self):
        pass

    # ---------------------- Stats Logger ----------------------
    def getStatistics(self, start_time):
        print("Total Execution Time:", time.time() - start_time)
        process = psutil.Process()
        memory_kb = process.memory_full_info().uss / 1024
        print("Memory Usage (KB):", memory_kb)

    # ---------------------- One-Class SVM (sequential Sample Loop) ----------------------
    def compute_ocsvm_sequential(self, training_np, testing_np):
        print("Training One-Class SVM (sequential)...")
        model = OneClassSVM(kernel='rbf', gamma='scale')
        model.fit(training_np)

        print("Scoring test samples (loop)...")
        scores = np.array([model.decision_function([x])[0] for x in tqdm(testing_np, desc="Scoring (sequential)")])
        return scores

    # ---------------------- One-Class SVM (Vectorized Batch) ----------------------
    def compute_ocsvm_parallel(self, training_np, testing_np):
        print("Training One-Class SVM (parallel)...")
        model = OneClassSVM(kernel='rbf', gamma='scale')
        model.fit(training_np)

        print("Scoring test samples (batch)...")
        scores = model.decision_function(testing_np)
        return scores

    # ---------------------- Main Entry Point ----------------------
    def run(self, training, testing, topK=-1, mode="sequential", algorithm='OneClassSVM'):
        start_time = time.time()

        # Optional preprocessing
        if algorithm == "difOneClassSVM":
            training_df = training.diff(axis=1).iloc[:, 1:]
            testing = testing.diff(axis=1).iloc[:, 1:]

        # Convert to NumPy for faster processing
        training_np = training.to_numpy()
        testing_np = testing.to_numpy()

        # Scoring
        if mode == "sequential":
            scores = self.compute_ocsvm_sequential(training_np, testing_np)
        elif mode == "parallel":
            scores = self.compute_ocsvm_parallel(training_np, testing_np)
        else:
            raise ValueError("Invalid mode. Choose 'sequential' or 'parallel'")

        # Return DataFrame with scores and sorted top elements
        testing = testing.copy()
        testing['OCSVM_Score'] = scores
        sorted_df = testing.sort_values('OCSVM_Score', ascending=False).head(topK)

        self.getStatistics(start_time)
        return testing, sorted_df



import pandas as pd
import numpy as np
import time
import psutil
from tqdm import tqdm
from sklearn.ensemble import IsolationForest


class OneNNIsolationForest:

    def __init__(self):
        pass

    # ---------------------- Stats Logger ----------------------
    def getStatistics(self, start_time):
        print("Total Execution Time:", time.time() - start_time)
        process = psutil.Process()
        memory_kb = process.memory_full_info().uss / 1024
        print("Memory Usage (KB):", memory_kb)

    # ---------------------- Isolation Forest Modes ----------------------
    def compute_iforest_sequential(self, training, testing):
        print("Training Isolation Forest (sequential)...")
        model = IsolationForest(n_estimators=100, contamination='auto', n_jobs=1, random_state=42)
        model.fit(training)

        print("Scoring test samples...")
        scores = model.decision_function(testing)
        return scores

    def compute_iforest_parallel(self, training, testing):
        print("Training Isolation Forest (parallel)...")
        model = IsolationForest(n_estimators=100, contamination='auto', n_jobs=-1, random_state=42)
        model.fit(training)

        print("Scoring test samples in parallel...")
        scores = model.decision_function(testing)
        return scores

    # ---------------------- Main Function ----------------------
    def run(self, training, testing, topK=-1, mode="sequential", algorithm='rasterIsolationForest'):
        start_time = time.time()

        if algorithm == "difrasterIsolationForest":
            training = training.diff(axis=1).iloc[:, 1:]
            testing = testing.diff(axis=1).iloc[:, 1:]

        if mode == "sequential":
            scores = self.compute_iforest_sequential(training, testing)
        elif mode == "parallel":
            scores = self.compute_iforest_parallel(training, testing)
        else:
            raise ValueError("Invalid mode. Choose 'sequential', or 'parallel'")

        testing['IForest_Score'] = scores
        sorted_df = testing.sort_values('IForest_Score', ascending=False).head(topK)
        self.getStatistics(start_time)
        return testing, sorted_df


"""
Exact GPU-based 1-NN Classifiers + One-Class Models (Single, Parallel, CUDA)
Includes: Manhattan, One-Class SVM, Isolation Forest, K-Means, Autoencoder (Improved with Batch Scoring)
"""

import pandas as pd
import numpy as np
import time
import psutil
from tqdm import tqdm
from sklearn.preprocessing import StandardScaler
import torch
import torch.nn as nn
import torch.optim as optim
from torch.utils.data import DataLoader, TensorDataset

torch.device('cuda' if torch.cuda.is_available() else 'cpu')

def getStatistics(start_time):
    print("Total Execution Time:", time.time() - start_time)
    process = psutil.Process()
    memory_kb = process.memory_full_info().uss / 1024
    print("Memory Usage (KB):", memory_kb)

# ---------------------- Improved Autoencoder Model ----------------------
class Autoencoder(nn.Module):
    def __init__(self, input_dim):
        super(Autoencoder, self).__init__()
        self.encoder = nn.Sequential(
            nn.Linear(input_dim, 128),
            nn.ReLU(),
            nn.Linear(128, 64),
            nn.ReLU(),
            nn.Linear(64, 16)
        )
        self.decoder = nn.Sequential(
            nn.Linear(16, 64),
            nn.ReLU(),
            nn.Linear(64, 128),
            nn.ReLU(),
            nn.Linear(128, input_dim),
            nn.Sigmoid()
        )

    def forward(self, x):
        encoded = self.encoder(x)
        decoded = self.decoder(encoded)
        return decoded

# ---------------------- Autoencoder Modes ----------------------
def compute_autoencoder(training, testing, mode="cuda", epochs=100, batch_size=64):
    device = torch.device("cuda" if mode == "cuda" and torch.cuda.is_available() else "cpu")

    scaler = StandardScaler()
    training_np = scaler.fit_transform(training.to_numpy())
    testing_np = scaler.transform(testing.to_numpy())

    x_train = torch.tensor(training_np, dtype=torch.float32)
    x_test = torch.tensor(testing_np, dtype=torch.float32)

    train_loader = DataLoader(TensorDataset(x_train), batch_size=batch_size, shuffle=True)
    test_loader = DataLoader(TensorDataset(x_test), batch_size=batch_size, shuffle=False)

    model = Autoencoder(x_train.shape[1]).to(device)
    criterion = nn.MSELoss()
    optimizer = optim.Adam(model.parameters(), lr=1e-3)

    model.train()
    for epoch in range(epochs):
        total_loss = 0
        for batch in train_loader:
            batch_data = batch[0].to(device)
            optimizer.zero_grad()
            output = model(batch_data)
            loss = criterion(output, batch_data)
            loss.backward()
            optimizer.step()
            total_loss += loss.item() * batch_data.size(0)
        avg_loss = total_loss / len(x_train)
        print(f"Epoch {epoch+1}/{epochs}, Avg Loss: {avg_loss:.4f}")

    print("Scoring test samples in batches...")
    model.eval()
    errors = []
    with torch.no_grad():
        for batch in tqdm(test_loader, desc="Testing", leave=False):
            batch_data = batch[0].to(device)
            output = model(batch_data)
            loss = torch.mean((batch_data - output) ** 2, dim=1)
            errors.extend(loss.cpu().numpy())

    return np.array(errors)

# ---------------------- Main Entry ----------------------
def rasterOneClassAutoencoder(training, testing, topElements=-1):
    start_time = time.time()
    distances = compute_autoencoder(training, testing)
    testing['AE_ReconError'] = distances
    sorted_df = testing.sort_values('AE_ReconError').head(topElements)
    getStatistics(start_time)
    return testing, sorted_df



import pandas as pd
import numpy as np
import time
import psutil
from tqdm import tqdm
from numba import njit, prange, cuda

# ---------------------- Numba Hausdorff (Optimized) ----------------------
@njit
def hausdorff_numba(u, v):
    row = u.shape[0]
    lea_distance = 0.0
    for i in range(row):
        diffs = np.empty(row - 1)
        for k in range(row - 1):
            diffs[k] = abs((u[i] - v[k + 1]) - (u[i] - v[k]))
        distance1 = np.min(diffs)
        if distance1 > lea_distance:
            lea_distance = distance1
    return lea_distance

@njit
def hausdorff_distance(u, v):
    return max(hausdorff_numba(u, v), hausdorff_numba(v, u))

@njit
def compute_hausdorff_sequential(test_np, train_np):
    num_test = test_np.shape[0]
    num_train = train_np.shape[0]
    distances = np.empty(num_test)
    for i in range(num_test):
        min_dist = 1e10
        for j in range(num_train):
            dist = hausdorff_distance(test_np[i], train_np[j])
            if dist < min_dist:
                min_dist = dist
        distances[i] = min_dist
    return distances

@njit(parallel=True)
def compute_hausdorff_parallel(test_np, train_np):
    num_test = test_np.shape[0]
    num_train = train_np.shape[0]
    distances = np.empty(num_test)
    for i in prange(num_test):
        min_dist = 1e10
        for j in range(num_train):
            dist = hausdorff_distance(test_np[i], train_np[j])
            if dist < min_dist:
                min_dist = dist
        distances[i] = min_dist
    return distances

# ---------------------- Class Wrapper ----------------------
class OneNNHausdorff:
    def get_statistics(self, start_time):
        print("Total Execution Time:", time.time() - start_time)
        process = psutil.Process()
        memory_kb = process.memory_full_info().uss / 1024
        print("Memory Usage (KB):", memory_kb)

    @staticmethod
    @cuda.jit
    def custom_hausdorff_kernel(testing, training, result):
        test_idx = cuda.grid(1)
        if test_idx >= testing.shape[0]:
            return

        num_features = testing.shape[1]
        min_total_dist = 1e10

        for train_idx in range(training.shape[0]):
            max_min1 = 0.0
            for i in range(num_features):
                min_diff = 1e10
                for k in range(num_features - 1):
                    diff = abs((testing[test_idx][i] - training[train_idx][k + 1]) -
                               (testing[test_idx][i] - training[train_idx][k]))
                    if diff < min_diff:
                        min_diff = diff
                if min_diff > max_min1:
                    max_min1 = min_diff

            max_min2 = 0.0
            for i in range(num_features):
                min_diff = 1e10
                for k in range(num_features - 1):
                    diff = abs((training[train_idx][i] - testing[test_idx][k + 1]) -
                               (training[train_idx][i] - testing[test_idx][k]))
                    if diff < min_diff:
                        min_diff = diff
                if min_diff > max_min2:
                    max_min2 = min_diff

            final_dist = max(max_min1, max_min2)
            if final_dist < min_total_dist:
                min_total_dist = final_dist

        result[test_idx] = min_total_dist

    def compute_hausdorff_cuda(self, testing, training):
        test_np = testing.to_numpy().astype(np.float64)
        train_np = training.to_numpy().astype(np.float64)
        result = cuda.device_array(test_np.shape[0], dtype=np.float64)

        threads_per_block = 128
        blocks_per_grid = (test_np.shape[0] + threads_per_block - 1) // threads_per_block
        self.custom_hausdorff_kernel[blocks_per_grid, threads_per_block](test_np, train_np, result)

        return result.copy_to_host()

    def run(self, training, testing, topK=-1, mode="sequential", algorithm='Hausdorff'):
        start_time = time.time()

        if algorithm == "difHausdorff":
            training = training.diff(axis=1).iloc[:, 1:]
            testing = testing.diff(axis=1).iloc[:, 1:]

        test_np = testing.to_numpy()
        train_np = training.to_numpy()

        if mode == "sequential":
            distances = compute_hausdorff_sequential(test_np, train_np)
        elif mode == "parallel":
            distances = compute_hausdorff_parallel(test_np, train_np)
        elif mode == "cuda":
            if not cuda.is_available():
                raise RuntimeError("CUDA is not available on this machine.")
            distances = self.compute_hausdorff_cuda(testing, training)
        else:
            raise ValueError("Invalid mode. Choose 'sequential', 'parallel', or 'cuda'")

        testing['1NNHausdorff'] = distances
        sorted_df = testing.sort_values('1NNHausdorff').head(topK)
        self.get_statistics(start_time)
        return testing, sorted_df


import pandas as pd
from tqdm import tqdm
from numba import njit, prange, cuda
import numpy as np
import time
import gc
import os
import psutil
import sys


class FuzzyTSC:
    def __init__(self):
        pass

    def getStatistics(self, startTime):
        """Statistics of the Algorithm after execution"""
        print("Total Execution time of proposedAlgo:", time.time() - startTime)
        process = psutil.Process(os.getpid())
        memory = process.memory_full_info().uss
        memory_in_KB = memory / 1024
        print("Memory of proposedAlgo in KB:", memory_in_KB)

    @staticmethod
    def compute_rd_sequential(testing_array, mean_data, min_data, max_data, num_columns):
        """Compute RD scores for sequential-threaded execution."""
        num_rows = testing_array.shape[0]
        rd_scores = np.zeros(num_rows)

        for i in tqdm(range(num_rows)):
            counter = 0.0
            for j in range(1, num_columns):  # Skip the first column
                value = testing_array[i, j]
                if value >= mean_data[j]:
                    if value <= max_data[j]:
                        counter += 0.5 * (value - mean_data[j]) / (max_data[j] - mean_data[j])
                    else:
                        counter += 1
                else:
                    if value >= min_data[j]:
                        counter += 0.5 * (mean_data[j] - value) / (mean_data[j] - min_data[j])
                    else:
                        counter += 1
            rd_scores[i] = counter / num_columns

        return rd_scores

    @staticmethod
    @njit(parallel=True)
    def compute_rd_parallel(testing_array, mean_data, min_data, max_data, num_columns):
        """Compute RD scores using multi-threaded parallel processing with Numba."""
        num_rows = testing_array.shape[0]
        rd_scores = np.zeros(num_rows)

        for i in prange(num_rows):
            counter = 0.0
            for j in range(1, num_columns):  # Skip the first column
                value = testing_array[i, j]
                if value >= mean_data[j]:
                    if value <= max_data[j]:
                        counter += 0.5 * (value - mean_data[j]) / (max_data[j] - mean_data[j])
                    else:
                        counter += 1
                else:
                    if value >= min_data[j]:
                        counter += 0.5 * (mean_data[j] - value) / (mean_data[j] - min_data[j])
                    else:
                        counter += 1
            rd_scores[i] = counter / num_columns

        return rd_scores

    @staticmethod
    @cuda.jit
    def compute_rd_kernel(testing_array, mean_data, min_data, max_data, rd_scores, num_columns):
        """CUDA kernel for RD score computation."""
        i = cuda.grid(1)
        if i < testing_array.shape[0]:  # Ensure index is within bounds
            counter = 0.0
            for j in range(1, num_columns):  # Skip the first column
                value = testing_array[i, j]
                if value >= mean_data[j]:
                    if value <= max_data[j]:
                        counter += 0.5 * (value - mean_data[j]) / (max_data[j] - mean_data[j])
                    else:
                        counter += 1
                else:
                    if value >= min_data[j]:
                        counter += 0.5 * (mean_data[j] - value) / (mean_data[j] - min_data[j])
                    else:
                        counter += 1
            rd_scores[i] = counter / num_columns

    def compute_rd_cuda(self, testing_array, mean_data, min_data, max_data, num_columns):
        """Compute RD scores using CUDA for GPU acceleration."""
        num_rows = testing_array.shape[0]

        # Allocate GPU memory
        d_testing_array = cuda.to_device(testing_array)
        d_mean_data = cuda.to_device(mean_data)
        d_min_data = cuda.to_device(min_data)
        d_max_data = cuda.to_device(max_data)
        d_rd_scores = cuda.device_array(num_rows, dtype=np.float32)

        # Define thread and block configuration
        threads_per_block = 256
        blocks_per_grid = (num_rows + (threads_per_block - 1)) // threads_per_block

        # Launch CUDA kernel
        self.compute_rd_kernel[blocks_per_grid, threads_per_block](
            d_testing_array, d_mean_data, d_min_data, d_max_data, d_rd_scores, num_columns
        )

        # Copy RD scores back to host
        return d_rd_scores.copy_to_host()


    def run(self, training, testing, topK=-1, mode="sequential", algorithm="FuzzyTSC"):
        start_time = time.time()

        if algorithm == "difFuzzyTSC":
            training = training.diff(axis=1).iloc[:, 1:]
            testing = testing.diff(axis=1).iloc[:, 1:]

        # Calculate max, mean, and min values for each feature in the training dataset
        max_data = training.max(axis=0).values
        mean_data = training.mean(axis=0).values
        min_data = training.min(axis=0).values

        # Clean up training data
        del training
        gc.collect()

        # Convert testing dataset to numpy array for compatibility with Numba and CUDA
        testing_array = testing.to_numpy()
        num_rows, num_columns = testing_array.shape

        if mode == "sequential":
            rd_scores = self.compute_rd_sequential(testing_array, mean_data, min_data, max_data, num_columns)
        elif mode == "parallel":
            rd_scores = self.compute_rd_parallel(testing_array, mean_data, min_data, max_data, num_columns)
        elif mode == "cuda":
            if not cuda.is_available():
                raise RuntimeError("CUDA is not available on this machine.")
            rd_scores = self.compute_rd_cuda(testing_array, mean_data, min_data, max_data, num_columns)
        else:
            raise ValueError("Invalid mode. Choose 'sequential', 'parallel', or 'cuda'.")

        # Add RD scores to the testing DataFrame
        testing['RD'] = rd_scores

        # Retrieve top elements based on RD scores
        sorted_df = testing.sort_values('RD').head(topK)

        # Log statistics
        self.getStatistics(start_time)

        return testing, sorted_df



import pandas as pd
import numpy as np
import time
import math
import os
import psutil
from tqdm import tqdm
from numba import njit, prange, cuda


class OneNNED:
    def __init__(self):
        pass

    @staticmethod
    def get_statistics(start_time):
        print("Total Execution Time:", time.time() - start_time)
        process = psutil.Process(os.getpid())
        memory_kb = process.memory_full_info().uss / 1024
        print("Memory Usage (KB):", memory_kb)

    # ---------------------- Optimized sequential Thread (NumPy Vectorization) ----------------------
    @staticmethod
    def compute_ed_sequential(testing, training):
        test_np = testing.to_numpy()
        train_np = training.to_numpy()

        # Ensure both have the same number of columns (features)
        if test_np.shape[1] != train_np.shape[1]:
            min_features = min(test_np.shape[1], train_np.shape[1])
            test_np = test_np[:, :min_features]
            train_np = train_np[:, :min_features]

        # Compute pairwise squared differences
        dists = np.sqrt(((test_np[:, np.newaxis, :] - train_np[np.newaxis, :, :]) ** 2).sum(axis=2))

        # Take the minimum distance for each test sample
        return np.min(dists, axis=1).tolist()

    # ---------------------- Parallel Version with Numba ----------------------
    @staticmethod
    @njit(parallel=True)
    def compute_ed_parallel(test_np, train_np):
        num_test = test_np.shape[0]
        num_train = train_np.shape[0]
        num_features = train_np.shape[1]
        distances = np.empty(num_test)

        for i in prange(num_test):
            min_dist = 1e10
            for j in range(num_train):
                sq = 0.0
                for k in range(num_features):
                    diff = test_np[i, k] - train_np[j, k]
                    sq += diff * diff
                dist = math.sqrt(sq)
                if dist < min_dist:
                    min_dist = dist
            distances[i] = min_dist

        return distances

    # ---------------------- CUDA Kernel ----------------------
    @staticmethod
    @cuda.jit
    def compute_ed_cuda_kernel(test_data, train_data, result):
        i = cuda.grid(1)
        if i < test_data.shape[0]:
            min_dist = 1e10
            for j in range(train_data.shape[0]):
                sq = 0.0
                for k in range(test_data.shape[1]):
                    diff = test_data[i][k] - train_data[j][k]
                    sq += diff * diff
                dist = math.sqrt(sq)
                if dist < min_dist:
                    min_dist = dist
            result[i] = min_dist

    @staticmethod
    def compute_ed_cuda(testing, training):
        test_np = testing.to_numpy().astype(np.float32)
        train_np = training.to_numpy().astype(np.float32)
        result = cuda.device_array(test_np.shape[0], dtype=np.float32)

        threads_per_block = 128
        blocks_per_grid = (test_np.shape[0] + threads_per_block - 1) // threads_per_block

        OneNNED.compute_ed_cuda_kernel[blocks_per_grid, threads_per_block](test_np, train_np, result)
        return result.copy_to_host()

    # ---------------------- Main Function ----------------------
    def run(self, training, testing, topK=-1, mode="sequential", algorithm='ED'):
        start_time = time.time()

        if algorithm == "difED":
            training = training.diff(axis=1).iloc[:, 1:]
            testing = testing.diff(axis=1).iloc[:, 1:]

        if mode == "sequential":
            distances = self.compute_ed_sequential(testing, training)
        elif mode == "parallel":
            distances = self.compute_ed_parallel(testing.to_numpy(), training.to_numpy())
        elif mode == "cuda":
            if not cuda.is_available():
                raise RuntimeError("CUDA is not available on this machine.")
            distances = self.compute_ed_cuda(testing, training)
        else:
            raise ValueError("Invalid mode. Choose 'sequential', 'parallel', or 'cuda'")

        testing['1NNED'] = distances
        sorted_df = testing.sort_values('1NNED').head(topK)
        self.get_statistics(start_time)

        return testing, sorted_df


class ScoreCalculator:
    def __init__(self, min_val, max_val, avg_val):
        self.min_val = min_val
        self.max_val = max_val
        self.avg_val = avg_val

    def calculate_score(self, value):
        if value == self.avg_val:
            return 0
        elif value == self.min_val or value == self.max_val:
            return 0.5
        elif value > self.max_val or value < self.min_val:
            return 1
        elif value < self.avg_val:
            # Linear interpolation between min_val and avg_val
            return 0.5 * (1 - (value - self.min_val) / (self.avg_val - self.min_val))
        else:
            # Linear interpolation between avg_val and max_val
            return 0.5 * (1 - (self.max_val - value) / (self.max_val - self.avg_val))
        



from sklearn.cluster import KMeans as kmeansAlg
from sklearn.cluster import DBSCAN
from sklearn.cluster import Birch
from sklearn.cluster import AgglomerativeClustering
from sklearn.cluster import MeanShift
from sklearn.cluster import OPTICS
from sklearn.cluster import SpectralClustering
from sklearn.cluster import AffinityPropagation
from sklearn.preprocessing import StandardScaler
from sklearn.mixture import GaussianMixture
from numpy import unique
import psycopg2
import numpy as np
import pandas as pd


def kMeans(dataframe, k, max_iter=100):
    """
    :Description:
           Kmeans algorithm is an iterative algorithm that tries to partition the dataset into K pre-defined distinct
           non-overlapping subgroups (clusters) where each data point belongs to only one group.

    :param dataframe: Input data frame containing numerical data to be clustered.
    :param k: number of clusters
    :param max_iter: maximum number of iterations

    :return: Index of the cluster each sample belongs to and cluster centers
    :rtype: ndarray of shape (n_samples,)
    """
    # initialize k-means algorithm
    data = dataframe.drop(['x', 'y'], axis=1)
    data = data.to_numpy()
    kmeans = kmeansAlg(n_clusters=k, max_iter=max_iter).fit(data)
    # return the clusters
    # combine x ,y and labels as dataframe
    label = dataframe[['x', 'y']]
    labels = label.assign(labels=kmeans.labels_)

    return labels, kmeans.cluster_centers_


def kMeansPP(dataframe, k, max_iter=300):
    """
    :Description:
           To overcome the  drawbacks of standard k-means algorithm we use K-means++. This algorithm ensures a smarter
           initialization of the centroids and improves the quality of the clustering.

    :param dataframe: Input data frame containing numerical data to be clustered.
    :param k: number of clusters
    :param max_iter: maximum number of iterations

    :return: Index of the cluster each sample belongs to and cluster centers
    :rtype: ndarray of shape (n_samples,)
    """
    # initialize k-means++ algorithm
    data = dataframe.drop(['x', 'y'], axis=1)
    data = data.to_numpy()
    kmeans = kmeansAlg(n_clusters=k, max_iter=max_iter, init='k-means++').fit(data)
    # return the clusters
    # combine x ,y and labels as dataframe
    label = dataframe[['x', 'y']]
    labels = label.assign(labels=kmeans.labels_)

    return labels, kmeans.cluster_centers_


def DBScan(dataframe, ep, min_sample):
    """
    :Description:
        DBSCAN is a density-based clustering algorithm that groups together data points that are close to each other
        based on the specified parameters 'eps' and 'min_samples'. The algorithm does not require the number of clusters
        to be specified in advance and can discover clusters of arbitrary shapes.

    :param dataframe: Input data frame containing numerical data to be clustered.

    :param ep (float): The maximum distance between two samples for them to be considered as in the same neighborhood.

    :param min_sample (int): The number of samples (data points) in a neighborhood for a point to be considered as a core point.

    :return: Index of the cluster each sample belongs to.
    :rtype: ndarray of shape (n_samples,)

    """


    # initialize k-means++ algorithm
    data = dataframe.drop(['x', 'y'], axis=1)
    data = data.to_numpy()
    dbs = DBSCAN(eps=ep, min_samples=min_sample).fit(data)
    # return the clusters
    # combine x ,y and labels as dataframe
    label = dataframe[['x', 'y']]
    labels = label.assign(labels=dbs.labels_)

    return labels


def affinityPropagation(dataframe, damping_factor, max_iter=300, convergence_iter=15, affinity='euclidean',
                        random_state=None, preference=None):
    """
    :Description:
       Affinity Propagation is a clustering algorithm used to group data points into clusters without requiring
       the number of clusters as an input. It uses a message-passing approach to determine the exemplars (representative points) for each cluster. The exemplars are data points that best represent the cluster.

    :param dataframe: data to be clustered
    :param damping factor: In the range [0.5, 1.0) -o avoid numerical oscillations when updating values
    :param max_iter: maximum number of iterations
    :param convergence_iter: default = 15
    :param affinity: {‘euclidean’, ‘precomputed’}
    :param random_state: int, RandomState instance or None, default=None
    :param preference: default = None :points with larger values of preferences are more likely to be chosen as exemplars

    :return: Index of the cluster each sample belongs to and cluster centers
    :rtype: ndarray of shape (n_samples,)
    """
    # initialize affinity propagation algorithm
    data = dataframe.drop(['x', 'y'], axis=1)
    data = data.to_numpy()
    X = StandardScaler().fit_transform(data)
    affinityProp = AffinityPropagation(damping=float(damping_factor), max_iter=int(max_iter),
                                       convergence_iter=int(convergence_iter), affinity=affinity, preference=preference,
                                       random_state=random_state).fit(X)
    # return the clusters
    # combine x ,y and labels as dataframe
    label = dataframe[['x', 'y']]
    labels = label.assign(labels=affinityProp.labels_)

    return labels, affinityProp.cluster_centers_


def BIRCH(dataframe, threshold, branch_factor=50, compute_labels=True, n_clusters=3):
    """
    :Description: BIRCH - (Balanced Iterative Reducing and Clustering using Hierarchies) involves constructing a tree structure from which cluster centroids are extracted
    :param dataframe: data to be clustered
    :param threshold: (default = 0.5) The radius of the subcluster obtained by merging a new sample and the closest subcluster should be lesser than the threshold.
    :param branching_factor: (default = 50) Maximum number of CF subclusters in each node.
    :param n_clusters: n_clustersint, instance of sklearn.cluster model, default=3
    :param compute_labels: True .Whether or not to compute labels for each fit
    :return: Index of the cluster each sample belongs to and cluster centers
    :rtype: ndarray of shape (n_samples,)
    """
    # initialize affinity propagation algorithm
    data = dataframe.drop(['x', 'y'], axis=1)
    data = np.ascontiguousarray(data.to_numpy())

    birch = Birch(n_clusters=int(n_clusters), threshold=float(threshold), branching_factor=int(branch_factor),
                  compute_labels=bool(branch_factor)).fit(data)

    # return the clusters
    # combine x ,y and labels as dataframe
    label = dataframe[['x', 'y']]
    labels = label.assign(labels=birch.labels_)

    return labels, birch.subcluster_centers_


def agglomerativeClustering(dataframe, n_clusters=2):
    """
    :Description: Agglomerative Clustering - Recursively merges pair of clusters of sample data; uses linkage distance

    :param dataframe: data to be clustered
    :param n_clusters: The number of clusters to find.

    :return: Index of the cluster each sample belongs to.
    :rtype: ndarray of shape (n_samples,)
    """
    # initialize affinity propagation algorithm
    data = dataframe.drop(['x', 'y'], axis=1)
    data = data.to_numpy()
    # X = StandardScaler().fit_transform(data)

    agglomerativeClustering = AgglomerativeClustering(n_clusters=n_clusters).fit(data)

    # return the clusters
    # combine x ,y and labels as dataframe
    label = dataframe[['x', 'y']]
    labels = label.assign(labels=agglomerativeClustering.labels_)

    return labels


def meanShift(dataframe, bandwidth=None, max_iters=300):
    """
    :Description: Mean shift clustering aims to discover “blobs” in a smooth density of samples.
    :param dataframe: input data frame containing data to be clustered
    :param bandwidth: float, default=None
    :param max_iters: maximum number of iterations, default=300
    :return: Index of the cluster each sample belongs to and cluster centers
    :rtype: ndarray of shape (n_samples,)
    """
    # initialize affinity propagation algorithm
    data = dataframe.drop(['x', 'y'], axis=1)
    data = data.to_numpy()
    # X = StandardScaler().fit_transform(data)

    meanShift = MeanShift(max_iter=max_iters).fit(data)
    # combine x ,y and labels as dataframe
    label = dataframe[['x', 'y']]
    labels = label.assign(labels=meanShift.labels_)

    return labels, meanShift.cluster_centers_


def opticsClustering(dataframe, min_samples=5, eps=None):
    """
    :Description: OPTICS, closely related to DBSCAN, finds core sample of high density and expands clusters from them
    :param dataframe: data to be clustered
    :param min_samples: number of samples in a neighborhood for a point to be considered as a core point.
    :param eps: maximum distance between two samples for one to be considered as in the neighborhood of the other
    :return: Index of the cluster each sample belongs to.
    :rtype: ndarray of shape (n_samples,)
    """
    # initialize affinity propagation algorithm
    data = np.ascontiguousarray(np.array(dataframe.drop(['x', 'y'], axis=1)))

    OPTICS_Clustering = OPTICS(min_samples=min_samples, eps=eps).fit(data)
    # combine x ,y and labels as dataframe
    label = dataframe[['x', 'y']]
    labels = label.assign(labels=OPTICS_Clustering.labels_)

    return labels


def spectralClustering(dataframe, n_clusters=8, assign_labels='discretize'):
    """
    :Description: Spectral Clustering is useful when the structure of the individual clusters is highly non-convex
    :param dataframe: data to be clustered
    :param n_clusters: The dimension of the projection subspace
    :param assign_labels: {‘kmeans’, ‘discretize’, ‘cluster_qr’}, default=’kmeans’
    :return: Index of the cluster each sample belongs to
    :rtype: ndarray of shape (n_samples,)
    """
    # initialize affinity propagation algorithm
    data = np.ascontiguousarray(np.array(dataframe.drop(['x', 'y'], axis=1)))

    spectralClustering = SpectralClustering(assign_labels=assign_labels, n_clusters=n_clusters, random_state=0).fit(
        data)
    # combine x ,y and labels as dataframe
    label = dataframe[['x', 'y']]
    labels = label.assign(labels=spectralClustering.labels_)

    return labels


def gaussianMixture(dataframe, n_components=1, max_iters=100, covariance_type="full", init_params='kmeans',
                    random_state=0):
    """
    :Description: To estimate the parameters of a Gaussian mixture distribution
    :param dataframe: data to be clustered
    :param max_iter: int, default=100
    :param random_state: int, RandomState instance or None, default=None
    :param covariance_type: {‘full’, ‘tied’, ‘diag’, ‘spherical’}, default=’full’
    :param init_params: {‘kmeans’, ‘k-means++’, ‘random’, ‘random_from_data’}
    :return: clusters,weights(weight of each mixture components, means (mean of each component)),
    """
    # initialize affinity propagation algorithm
    data = np.array(dataframe.drop(['x', 'y'], axis=1))
    gaussianMixture = GaussianMixture(n_components=n_components, max_iter=max_iters, covariance_type=covariance_type,
                                      init_params=init_params, random_state=random_state).fit(data)
    # assign each data point to a cluster
    gaussianResult = gaussianMixture.predict(data)
    label = dataframe[['x', 'y']]
    labels = label.assign(labels=gaussianResult)

    return labels, gaussianMixture.weights_, gaussianMixture.means_



#!/usr/bin/python
from configparser import ConfigParser
import os

def config(filename='database.ini', section='postgresql'):
    # create a parser
    parser = ConfigParser()
    # read config file
    parser.read(filename)

    # get section, default to postgresql
    db = {}
    if parser.has_section(section):
        params = parser.items(section)
        for param in params:
            db[param[0]] = param[1]
    else:
        raise Exception('Section {0} not found in the {1} file'.format(section, filename))

    return db


from geoAnalytics import csv2raster
from geoAnalytics.config import config
from shapely import geos, wkb, wkt
from osgeo import gdal
import numpy as Numeric
import pandas as pd
import subprocess
import psycopg2
import random
import glob
import sys
import os
import threading
from tqdm import tqdm
import multiprocessing
import time


def connect(dbName, hostIP, user, password, port=5432):
    """
    Connect to the database using the parameters.

    :param dbName: name of the database
    :param hostIP: host IP
    :param user: user name
    :param password: password
    :param port: port
    """

    if dbName != "":
        with open('database.ini', "w") as dbFile:
            dbFile.write("[postgresql]\nhost = " + str(hostIP) + "\nport = " + str(port) +
                         "\ndatabase = " + str(dbName) + "\nuser = " + str(user) + "\npassword = " + str(password))
            dbFile.close()
    connection = testConnection()


def disconnect():
    """
    Disconnect from the database

    """
    conn = None
    # read database configuration
    params = config()
    # connect to the PostgreSQL database
    conn = psycopg2.connect(**params)
    # create a new cursor
    curr = conn.cursor()
    if conn is not None:
        print("Disconnecting from repository")
        conn.close()
        print("Disconnected from repository")

# def reConnect(self, dbName="", hostIP="", user="", password="", port=5432):
#     """
#     Edit the connection to the database
#
#     :param dbName: name of the database
#     :param hostIP: host IP
#     :param user: user name
#     :param password: password
#     :param port: port
#     """
#     conn = None
#     if dbName != "":
#         with open('ini', "w") as dbFile:
#             buffer = "[postgresql]\nhost = " + hostIP + "\nport = " + str(
#                 port) + "\ndatabase = " + dbName + "\nuser = " + user + "\npassword = " + password
#             dbFile.write(buffer)
#             dbFile.close()
#     self.testConnection()


def testConnection():
    """
    Test the connection to the database
    """

    try:
        conn = None
        # read database configuration
        params = config()
        # connect to the PostgreSQL database
        conn = psycopg2.connect(**params)
        # create a new cursor
        curr = conn.cursor()
        curr.execute("select version();")
        for item in curr:
            print(item)
        print('You are now connected')
    except (Exception, psycopg2.DatabaseError) as error:
        print(error)


def create(repositoryName, totalBands, SRID=4326):
    """
    Create a repository in the database

    :param repositoryName: name of the repository
    :param totalBands: total number of bands
    :param SRID: spatial reference ID
    """

    try:
        conn = None
        # read database configuration
        params = config()
        # connect to the PostgreSQL database
        conn = psycopg2.connect(**params)
        # create a new cursor
        curr = conn.cursor()
        # total bands is number
        query = "geog geometry(POINT," + str(SRID) + "),"
        for i in range(1, totalBands):
            query += 'b' + str(i) + ' float,'
        query += 'b' + str(totalBands) + ' float'
        curr.execute("create table " + repositoryName + "(" + query + ")")
        curr.execute("CREATE INDEX index_" + repositoryName +
                     " ON " + repositoryName + " USING GIST (geog)")
        conn.commit()
        print('Repository created')
    except (Exception, psycopg2.DatabaseError) as error:
        print(error)
    finally:
        if conn is not None:
            conn.close()
            print('Repository connection closed.')


def delete(repositoryName):
    """
    Delete a repository from the database

    :param repositoryName: name of the repository
    """

    try:
        conn = None
        # read database configuration
        params = config()
        # connect to the PostgreSQL database
        conn = psycopg2.connect(**params)
        # create a new cursor
        curr = conn.cursor()
        curr.execute("drop table " + repositoryName + ";")
        conn.commit()
        print(repositoryName + ' deleted successfully')
    except (Exception, psycopg2.DatabaseError) as error:
        print(error)
    finally:
        if conn is not None:
            conn.close()
            print('Repository connection closed.')


def clone(repositoryName, cloneRepositoryName):
    """
    Clone a repository from the database

    :param repositoryName: name of the repository
    :param cloneRepositoryName: name of the cloned repository
    """
    try:
        conn = None
        # read database configuration
        params = config()
        # connect to the PostgreSQL database
        conn = psycopg2.connect(**params)
        # create a new cursor
        curr = conn.cursor()
        curr.execute("create table " + cloneRepositoryName +
                     " as (select * from " + repositoryName + ");")
        conn.commit()
        print('Repository cloned')
    except (Exception, psycopg2.DatabaseError) as error:
        print(error)
    finally:
        if conn is not None:
            conn.close()
            print('Repository connection closed.')


def rename(repositoryName, newRepositoryName):
    """
    Change the name of a repository in the database

    :param repositoryName: name of the repository
    :param newRepositoryName: new name of the repository
    """
    try:
        conn = None
        # read database configuration
        params = config()
        # connect to the PostgreSQL database
        conn = psycopg2.connect(**params)
        # create a new cursor
        curr = conn.cursor()
        curr.execute("ALTER TABLE IF EXISTS " +
                     repositoryName + " RENAME TO " + newRepositoryName + ";")
        conn.commit()
        print(repositoryName + ' changed to ' + newRepositoryName)
    except (Exception, psycopg2.DatabaseError) as error:
        print(error)
    finally:
        if conn is not None:
            conn.close()
            print('Repository connection closed.')


def insertRaster(repositoryName, fileName, totalBands, scalingFactor, SRID=4326):
    """
    Insert a TIFF file into the database

    :param repositoryName: name of the Repository
    :param fileName: name of the TIFF file
    :param totalBands: number of bands
    :param SRID: spatial reference ID
    :param scalingFactor: scaling factor
    """
    
    tempFile = _r2tsv(totalBands, fileName, scalingFactor, SRID)
    insertCSV(tempFile, repositoryName)
    if os.path.exists(tempFile):
        os.remove(tempFile)



def insertRasterFolder(repositoryName, folderName, totalBands, scalingFactor, extension=".lbl",SRID=4326, threads = 16):
    """
    Insert a TIFF file into the database

    :param repositoryName: name of the Repository
    :param fileName: name of the TIFF file
    :param totalBands: number of bands
    :param SRID: spatial reference ID
    :param scalingFactor: scaling factor
    :param threads: number of threads to spawn
    """
    files = []
    for file in os.listdir(folderName):
        if file.endswith(extension):
            files.append(file)

    print(files)

    # single thread
    # for i in tqdm(range(len(files))):
    #     tempFile = _r2tsv(totalBands, folderName + '/' + files[i], scalingFactor, SRID)
    #     insertCSV(tempFile, repositoryName)
    #     if os.path.exists(tempFile):
    #         os.remove(tempFile)

    # multiprocessing 
    
    pbar = tqdm(total=len(files))
    pool = multiprocessing.Pool(threads)
#     start_time = time.perf_counter()
    
    def update(*a):
        pbar.update()
    for i in range(pbar.total):
        pool.apply_async(insertRaster, args=(repositoryName, folderName + '/' + files[i], totalBands, scalingFactor, SRID),callback=update)
    pool.close()
    pool.join()
    # processes = [pool.apply_async(insertRaster, args=(repositoryName, folderName + '/' + file, totalBands, scalingFactor, SRID)) for file in files]
    # result = [p.get() for p in processes]
#     finish_time = time.perf_counter()
#     print(f"Program finished in {finish_time-start_time} seconds")

def insertCSV(filename, repositoryName, seperator=' '):
    """
    Insert a CSV file into the database

    :param filename: name of the CSV file
    :param repositoryName: name of the Repository
    :param seperator: seperator
    """

    # global metadata
    try:
        conn = None
        # read database configuration
        params = config()
        # connect to the PostgreSQL database
        conn = psycopg2.connect(**params)
        # create a new cursor
        curr = conn.cursor()
        buffer = "COPY " + str(repositoryName) + " FROM '" + str(os.getcwd()) + '/' + str(
            filename) + "' DELIMITER '" + str(seperator) + "' CSV;"
        curr.execute(buffer)
        conn.commit()

#         print('File inserted')
    except (Exception, psycopg2.DatabaseError) as error:
        print(error)
    finally:
        if conn is not None:
            conn.close()

    # return gdal.Open(filename).GetMetadata()
#             print('Repository connection closed.')



def _r2tsv(endBand, srcfile, scalingFactor, SRID):
    """
    Convert a raster to a tsv file

    :param startBand: start band
    :param endBand: end band
    :param srcfile: source file
    :param dstfile: destination file
    :param scalingFactor: scaling factor
    """

    tempFile = str(random.randint(0, 100000)) + ".txt"
    band_nums = range(1, endBand + 1)
    srcwin = None
    if band_nums == []:
        band_nums = [1]

    # Open source file.
    srcds = gdal.Open(srcfile)
    if srcds is None:
        print('Could not open %s.' % srcfile)
        sys.exit(1)

    bands = []
    for band_num in band_nums:
        band = srcds.GetRasterBand(band_num)
        if band is None:
            print('Could not get band %d' % band_num)
            sys.exit(1)
        bands.append(band)
    gt = srcds.GetGeoTransform()

    # Collect information on all the source files.
    if srcwin is None:
        srcwin = (0, 0, srcds.RasterXSize, srcds.RasterYSize)

    # Open the output file.
    if tempFile is not None:
        dst_fh = open(tempFile, 'wt')
    else:
        dst_fh = sys.stdout
    band_format = ("%g " * len(bands)).rstrip() + '\n'

    # Setup an appropriate print format.
    if abs(gt[0]) < 180 and abs(gt[3]) < 180 \
            and abs(srcds.RasterXSize * gt[1]) < 180 \
            and abs(srcds.RasterYSize * gt[5]) < 180:
        format = '%s %s'
    else:
        format = '%s %s'

    # Loop emitting data.
    for y in range(srcwin[1], srcwin[1] + srcwin[3]):
        data = []
        for band in bands:
            band_data = band.ReadAsArray(srcwin[0], y, srcwin[2], 1)
            band_data = Numeric.reshape(band_data, (srcwin[2],))
            data.append(band_data)

        for x_i in range(0, srcwin[2]):
            x = x_i + srcwin[0]
            geo_x = gt[0] + (x + 0.5) * gt[1] + (y + 0.5) * gt[2]
            geo_y = gt[3] + (x + 0.5) * gt[4] + (y + 0.5) * gt[5]
            x_i_data = []
            for i in range(len(bands)):
                x_i_data.append(data[i][x_i]*scalingFactor)
            band_str = band_format % tuple(x_i_data)
            # line = format % (wkb.dumps(wkt.loads("POINT(" + str(float(geo_x)) + " " + str(float(geo_y)) + ")"),
            #                  hex=True, srid=SRID), band_str)  # Convert X and Y in hex, store the data, and upload
            line = format % (wkb.dumps(wkt.loads("POINT(" + str(round(geo_x, 4)) + " " + str(round(geo_y, 4)) + ")"),
                             hex=True, srid=SRID), band_str)  # Convert X and Y in hex, store the data, and upload
            dst_fh.write(line)
    return tempFile


def addBand(self, repositoryName, bandFormula):
    # Add a column to a table using alter Command
    try:
        conn = None
        # read database configuration
        params = config()
        # connect to the PostgreSQL database
        conn = psycopg2.connect(**params)
        # create a new cursor
        curr = conn.cursor()
        # describe table command

        curr.execute("ALTER TABLE " + repositoryName + " ADD COLUMN B" +
                     str(bandFormula) + " float" + ";")
        conn.commit()
        print('Band added to repository')
    except (Exception, psycopg2.DatabaseError) as error:
        print(error)
    finally:
        if conn is not None:
            conn.close()
            print('Repository connection closed.')


def deleteBand(repositoryName, bandNumber):
    # This function will delete the band number attribute from the table.
    try:
        conn = None
        # read database configuration
        params = config()
        # connect to the PostgreSQL database
        conn = psycopg2.connect(**params)
        # create a new cursor
        curr = conn.cursor()
        curr.execute("ALTER TABLE " + repositoryName + " DROP COLUMN B" +
                     str(bandNumber) + ";")
        conn.commit()
        print('Band deleted from repository')
    except (Exception, psycopg2.DatabaseError) as error:
        print(error)
    finally:
        if conn is not None:
            conn.close()
            print('Repository connection closed.')

    return


def getRaster(repositoryName, rasterFileName, Xmin, Ymin, Xmax, Ymax, Bands="*"):
    """
    Get a raster image from the database

    :param repositoryName: name of the repository
    :param rasterFileName: name of the raster file
    :param Xmin: minimum X coordinate
    :param Ymin: minimum Y coordinate
    :param Xmax: maximum X coordinate
    :param Ymax: maximum Y coordinate
    :param Bands: bands to be extracted
    """
    # connect to database
    # create geoTIFF file
    # gDal transform to geoTIFF
    print('Getting raster from database')
    dataFrame2Raster(getDataframeForEnvelope(
        repositoryName, Xmin, Ymin, Xmax, Ymax, Bands), rasterFileName)


def dataFrame2Raster(dataframe, rasterFileName):
    """
    Create a raster image from a dataframe

    :param dataframe: dataframe
    :param rasterFileName: name of the raster file with no file extension.

    """
    print('Creating raster file')

    df = dataframe
    csv2raster.csv2raster(output_file=rasterFileName, dataframe=df)


def getDataframe(repositoryName, Bands="*", SRID=4326):
    """
    Get a dataframe from the database for a given envelope

    :param repositoryName: name of the repository
    :param Xmin: minimum X coordinate
    :param Ymin: minimum Y coordinate
    :param Xmax: maximum X coordinate
    :param Ymax: maximum Y coordinate
    :param Bands: bands to be extracted
    :param SRID: spatial reference ID
    """

    # connect to database
    # create geoTIFF file
    # gDal transform to geoTIFF
    print('Getting dataframe from database')

    try:
        conn = None
        # read database configuration
        params = config()
        # connect to the PostgreSQL database
        conn = psycopg2.connect(**params)
        query = "SELECT ST_X(geog) as x, ST_Y(geog) as y, " + \
            Bands + " FROM " + str(repositoryName) + ";"
        dataFrameEnvelope = pd.read_sql_query(query, conn)
        print('Dataframe created')
        if Bands == "*":
            dataFrameEnvelope = dataFrameEnvelope.drop(columns=["geog"])
        return dataFrameEnvelope
    except (Exception, psycopg2.DatabaseError) as error:
        print(error)
    finally:
        if conn is not None:
            conn.close()
            print('Repository connection closed.')


# def getDataframe(repositoryName, Bands="*", SRID=4326):
#     """
#     Get a dataframe from the database for a given envelope
#
#     :param repositoryName: name of the repository
#     :param Bands: bands to be extracted
#     :param SRID: spatial reference ID
#     """
#
#     # connect to database
#     # create geoTIFF file
#     # gDal transform to geoTIFF
#     print('Getting dataframe from database')
#
#     try:
#         conn = None
#         # read database configuration
#         params = config()
#         # connect to the PostgreSQL database
#         conn = psycopg2.connect(**params)
#         query = "SELECT ST_X(geog) as x, ST_Y(geog) as y, " + \
#             Bands + " FROM " + str(repositoryName) + ";"
#         dataFrameEnvelope = pd.read_sql_query(query, conn)
#         print('Dataframe created')
#         if Bands == "*":
#             dataFrameEnvelope = dataFrameEnvelope.drop(columns=["geog"])
#         return dataFrameEnvelope
#     except (Exception, psycopg2.DatabaseError) as error:
#         print(error)
#     finally:
#         if conn is not None:
#             conn.close()
#             print('Repository connection closed.')


def getDataframeForEnvelope(repositoryName, Xmin, Ymin, Xmax, Ymax, Bands="*", SRID=4326):
    """
    Get a dataframe from the database for a given envelope

    :param repositoryName: name of the repository
    :param Xmin: minimum X coordinate
    :param Ymin: minimum Y coordinate
    :param Xmax: maximum X coordinate
    :param Ymax: maximum Y coordinate
    :param Bands: bands to be extracted
    :param SRID: spatial reference ID
    """

    # connect to database
    # create geoTIFF file
    # gDal transform to geoTIFF
    print('Getting dataframe from database')

    try:
        conn = None
        # read database configuration
        params = config()
        # connect to the PostgreSQL database
        conn = psycopg2.connect(**params)
        query = "SELECT ST_X(geog) as x, ST_Y(geog) as y, " + Bands + " FROM " + str(repositoryName) + " WHERE " + str(
            repositoryName) + ".geog && ST_MakeEnvelope(" + str(Xmin) + ',' + str(Ymin) + ',' + str(Xmax) + ',' + str(Ymax) + ");"
        dataFrameEnvelope = pd.read_sql_query(query, conn)
        print('Dataframe created')
        if Bands == "*":
            dataFrameEnvelope = dataFrameEnvelope.drop(columns=["geog"])
        return dataFrameEnvelope
    except (Exception, psycopg2.DatabaseError) as error:
        print(error)
    finally:
        if conn is not None:
            conn.close()
            print('Repository connection closed.')


def kNearestPixels(repositoryName, X, Y, k=1000, Bands="*", SRID=4326):
    """
    Get a dataframe from the database for a point and its neighbors

    :param repositoryName: name of the repository
    :param X: X coordinate
    :param Y: Y coordinate
    :param k: number of neighbours
    :param Bands: bands to be extracted
    :param SRID: spatial reference ID
    """
    print('Getting dataframe from database')

    try:
        conn = None
        # read database configuration
        params = config()
        # connect to the PostgreSQL database
        conn = psycopg2.connect(**params)
        query = "SELECT ST_X(geog) as x, ST_Y(geog) as y, " + str(Bands) + " FROM " + str(
            repositoryName) + " ORDER BY " + str(repositoryName) + ".geog <-> 'SRID=" + str(SRID) + ";POINT(" + str(
            X) + ' ' + str(Y) + ")' limit " + str(k) + ";"
        dataFrameEnvelope = pd.read_sql_query(query, conn)
        print('Dataframe created')
        if Bands == "*":
            dataFrameEnvelope = dataFrameEnvelope.drop(columns=["geog"])
        return dataFrameEnvelope
    except (Exception, psycopg2.DatabaseError) as error:
        print(error)
    finally:
        if conn is not None:
            conn.close()
            print('Repository connection closed.')


def getRasterForkNearestPixels(repositoryName, rasterFileName="rasterFile.nc", X=0, Y=0, k=1000, Bands="*"):
    """
    Get a raster image from the database for a point and its neighbors

    :param repositoryName: name of the repository
    :param rasterFileName: name of the raster file
    :param X: X coordinate
    :param Y: Y coordinate
    :param k: number of neighbours
    :param Bands: bands to be extracted
    """
    # connect to database
    # create geoTIFF file
    # gDal transform to geoTIFF
    df = kNearestPixels(repositoryName, X, Y, k, Bands)
    dataFrame2Raster(df, rasterFileName)


def getRepositorySize(tableName):
    """

    """

    try:
        conn = None
        # read database configuration
        params = config()
        # connect to the PostgreSQL database
        conn = psycopg2.connect(**params)
        query = "SELECT pg_size_pretty( pg_total_relation_size('" + \
            tableName + "') );"
        dataFrameEnvelope = pd.read_sql_query(query, conn)
        print(dataFrameEnvelope.iloc[0]['pg_size_pretty'])
        return dataFrameEnvelope.iloc[0]['pg_size_pretty']
    except (Exception, psycopg2.DatabaseError) as error:
        print(error)
    finally:
        if conn is not None:
            conn.close()
            print('Repository connection closed.')


def getSizeOfAllRepositories():
    """
    Get return the size of the repository/database/table

    :param repositoryName: name of the repository
    """

    try:
        conn = None
        # read database configuration
        params = config()
        # connect to the PostgreSQL database
        conn = psycopg2.connect(**params)
        q1 = "select current_database();"
        q2 = pd.read_sql_query(q1, conn)

        query = " SELECT pg_size_pretty( pg_database_size('" + \
            q2.iloc[0]['current_database'] + "') );"
        dataFrameEnvelope = pd.read_sql_query(query, conn)
        print(dataFrameEnvelope.iloc[0]['pg_size_pretty'])
        return dataFrameEnvelope.iloc[0]['pg_size_pretty']
    except (Exception, psycopg2.DatabaseError) as error:
        print(error)
    finally:
        if conn is not None:
            conn.close()
            print('Repository connection closed.')



import setuptools

with open('README.md', 'r') as fh:
    long_description = fh.read()

setuptools.setup(
    name='geoAnalytics',
    version='2025.06.04.2',
    author='Rage Uday Kiran',
    author_email='uday.rage@gmail.com',
    description='This software is being developed at the University of Aizu, Aizu-Wakamatsu, Fukushima, Japan',
    long_description=long_description,
    long_description_content_type='text/markdown',
    packages=setuptools.find_packages(),
    url='https://github.com/udayRage/geoAnalytics',
    license='GPLv3',
    install_requires=[            # All necessary packages utilized by our PAMI software
        'psutil',
        'pandas',
        'matplotlib',
        'resource',
        'validators',
        'urllib3',
        'psycopg2-binary',
        'shapely',
        'tqdm',
        'mplcursors',
        'sklearn'
    ],
    classifiers=[
        'Development Status :: 4 - Beta',      # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
        'Programming Language :: Python :: 3',
	'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.5',
)



from osgeo import gdal
import pandas as pd
import os
import random
import subprocess

class CSV2Raster:
    def __init__(self, input_file='', output_file='output.nc', sep=" ", dataframe=''):
        self.input_file = input_file
        self.inputfile_sep = sep
        self.tempOut = output_file
        self.output_file = 'output.nc'
        self.dataFrame = dataframe

    def convert(self):
        if self.input_file != '':
            self.dataFrame = pd.read_csv(self.input_file,
                                         sep=self.inputfile_sep,
                                         header=None, index_col=None)

        self.dataFrame = self.dataFrame.sort_values(['y', 'x'], ascending=[True, True])
        dataFrameColumns = self.dataFrame.columns

        randInt = str(random.randint(0, 100000))
        self.dataFrame = self.dataFrame.sort_values(by=['y', 'x'], ascending=[False, True])

        for i in range(2, len(dataFrameColumns)):
            self.dataFrame.to_csv("xyzformat.xyz", index=False, header=None, sep=" ")
            raster = gdal.Translate("temp_" + randInt + "_" + str(dataFrameColumns[i]) + ".nc", "xyzformat.xyz")
            self.dataFrame = self.dataFrame.drop(columns=dataFrameColumns[i])
            buffer = 'ncrename -v Band1,' + str(dataFrameColumns[i]) + " temp_" + randInt + "_" + str(dataFrameColumns[i]) + ".nc"
            print(subprocess.getstatusoutput(buffer))

        buffer = 'cdo -f nc2 cat ' + "temp_" + randInt + "_*.nc " + self.output_file
        print(subprocess.getstatusoutput(buffer))

        buffer = 'rm ' + "temp_" + randInt + "_*.nc"
        print(subprocess.getstatusoutput(buffer))

        if self.tempOut[-3:] == '.nc':
            os.rename(self.output_file, self.tempOut)
        elif self.tempOut[-3:] == 'iff' or self.tempOut[-3:] == 'tif':
            buffer = 'gdal_translate -of GTiff output.nc ' + self.tempOut
            print(subprocess.getstatusoutput(buffer))
            buffer = 'rm output.nc'
            print(subprocess.getstatusoutput(buffer))
        else:
            os.rename(self.output_file, self.tempOut + ".nc")

        os.remove("xyzformat.xyz")



# import required libraries
import pandas as pd
import time
import os
import psutil
import gc
import sys
import numpy as np
import math


def rasterFuzzyTSC(training, testing,topElements):
    startTime = time.time()

    # To drop the first column which is a class label of each sample
    # training.drop(0, axis='columns', inplace=True)

    # separating classes into different dictionary variables and creating mean, min and max curves
    maxData = training.max(axis=0)
    meanData = training.mean(axis=0)
    minData = training.min(axis=0)

    del training
    gc.collect()

    # Classification Phase

    counter = {}
    num_rows, num_columns = testing.shape

    # print(num_rows, num_columns)
    newColumn = list()
    # print(num_columns)
    correct = 0
    for i in range(num_rows):
        counter = 0.0
        for j in range(1, num_columns):
            if testing.iloc[i, j] >= meanData[j]:
                if testing.iloc[i, j] <= maxData[j]:
                    counter = counter + 0.5 * (testing.iloc[i, j] - meanData[j]) / (maxData[j] - meanData[j])
                else:
                    counter = counter + 1
            else:
                if testing.iloc[i, j] >= minData[j]:
                    counter = counter + 0.5 * (meanData[j] - testing.iloc[i, j]) / (meanData[j] - minData[j])
                else:
                    counter = counter + 1

        counter = counter / num_columns

        newColumn.append(counter)

    testing['RD'] = newColumn

    # N = int(input("Enter the how many number of top elements to be retrieved between: 1 to " + str(num_rows) + ":"))
    N = topElements
    sortedDF = testing.sort_values('RD').head(N)
    getStatistics(startTime)
    return testing, sortedDF


def rasterOneNNDTW(training, testing,topElements):
    startTime = time.time()
    # training.drop(0, axis='columns', inplace=True)
    num_rows_test, num_columns_test = testing.shape
    num_rows_train, num_columns_train = training.shape
    training_noclass = training  # self.training[:, 1:]
    testing_noclass = testing  # self.testing[:, 1:]

    newColumn = list()
    # predicted_label = None
    correct = 0

    for i in range(num_rows_test):
        # print(i)
        least_distance = float('inf')
        for j in range(num_rows_train):
            dist = dtw(testing_noclass.iloc[i].to_numpy(), training_noclass.iloc[j].to_numpy())
            if dist < least_distance:
                # predicted_label = self.training[j][0]
                least_distance = dist

        newColumn.append(least_distance)
        # if predicted_label == self.testing[i][0]:
        #    correct = correct + 1
    testing['1NNmaxNorm'] = newColumn   # print final test
    N = topElements
    sortedDF = testing.sort_values('1NNmaxNorm').head(N) # top k
    getStatistics(startTime)
    return testing,sortedDF


def rasterOneNNED(training, testing,topElements):
    startTime = time.time()
    # training.drop(0, axis='columns', inplace=True)
    num_rows_test, num_columns_test = testing.shape
    num_rows_train, num_columns_train = training.shape
    predicted_label = None
    correct = 0
    newColumn = list()

    for i in range(num_rows_test):
        least_distance = float('inf')
        for j in range(num_rows_train):
            squaring = 0
            for k in range(num_columns_train):
                squaring = squaring + (testing.iloc[i, k] - training.iloc[j, k]) ** 2
            dist = math.sqrt(squaring)

            if dist < least_distance:
                # predicted_label = self.training[j][0]
                least_distance = dist
        newColumn.append(least_distance)

    testing['1NNED'] = newColumn
    N = topElements
    sortedDF = testing.sort_values('1NNED').head(N)
    getStatistics(startTime)
    return testing,sortedDF



def rasterOneNNHausdorff(training, testing,topElements):
    startTime = time.time()
    # training.drop(0, axis='columns', inplace=True)
    num_rows_test, num_columns_test = testing.shape
    num_rows_train, num_columns_train = training.shape
    training_noclass = training  # self.training[:, 1:]
    testing_noclass = testing  # self.testing[:, 1:]
    # predicted_label = None
    # correct = 0
    newColumn = list()

    for i in range(num_rows_test):
        least_distance = float('inf')
        for j in range(num_rows_train):
            dist = max(hausdorff(testing_noclass.iloc[i].to_numpy(), training_noclass.iloc[j].to_numpy()),
                       hausdorff(training_noclass.iloc[j].to_numpy(),
                                      testing_noclass.iloc[i].to_numpy()))  # math.sqrt(squaring)
            if dist < least_distance:
                # predicted_label = self.training[j][0]
                least_distance = dist
        # if predicted_label == self.testing[i][0]:
        #    correct = correct + 1
        newColumn.append(least_distance)

    testing['1NNHausdorff'] = newColumn
    N = topElements
    sortedDF = testing.sort_values('1NNHausdorff').head(N)
    getStatistics(startTime)
    return testing,sortedDF

def rasterOneNNMaxNorm(training, testing,topElements):
    startTime = time.time()
    # training.drop(0, axis='columns', inplace=True)
    num_rows_test, num_columns_test = testing.shape
    num_rows_train, num_columns_train =     training.shape
    newColumn = list()

    for i in range(num_rows_test):
        least_distance = float('inf')
        for j in range(num_rows_train):
            maximum = float('-inf')
            for k in range(num_columns_train):
                temp = np.abs(testing.iloc[i, k] - training.iloc[j, k])
                if temp > maximum:
                    maximum = temp
            dist = maximum  # math.sqrt(squaring)
            if dist < least_distance:
                # predicted_label = self.training[j][0]
                least_distance = dist
        newColumn.append(least_distance)
        # if predicted_label == self.testing[i][0]:
        #    correct = correct + 1

    testing['1NNmaxNorm'] = newColumn
    N = topElements
    sortedDF = testing.sort_values('1NNmaxNorm').head(N)
    getStatistics(startTime)
    return testing,sortedDF

def hausdorff(u, v):
    row, = u.shape
    lea_distance = 0
    for i in range(row):
        distance1 = np.amin(np.absolute((np.diff(u[i] - v))))
        if distance1 > lea_distance:
            lea_distance = distance1
    return lea_distance


def dtw(A, B):
        N = len(A)
        M = len(B)
        d = np.zeros((N, M))
        for n in range(N):
            for m in range(M):
                d[n][m] = (A[n] - B[m]) ** 2
        D = np.zeros(d.shape)
        D[0][0] = d[0][0]
        for n in range(1, N):
            D[n][0] = d[n][0] + D[n - 1][0]
        for m in range(1, M):
            D[0][m] = d[0][m] + D[0][m - 1]
        for n in range(1, N):
            for m in range(1, M):
                D[n][m] = d[n][m] + min(D[n - 1][m], D[n - 1][m - 1], D[n][m - 1])
        return D[N - 1][M - 1]


def showFinalTestSamples(df):
    print("Final test samples after calculating the Ravi Distance (RD):")
    print(df)


def showFinalTopNSamples(df):
    print("Final top- N  testing samples retrieved based on the Ravi Distance:")
    print(df)


def saveFinalTestSamples(df, outputFile):
    df.iloc[:, -1].to_csv(outputFile, header=False)


def saveFinalTopNSamples(df, outputFile):
    df.iloc[:, -1].to_csv(outputFile, header=False)


def getStatistics(startTime):
    # Statistics of the Algorithm after execution
    # print("Dataset name:", "SampleRunTrainFile.txt")
    print("Total Execution time of proposedAlgo", time.time() - startTime)
    process = psutil.Process(os.getpid())
    memory = process.memory_full_info().uss
    memory_in_KB = memory / 1024
    print("Memory of proposedAlgo in KB:", memory_in_KB)  # in bytes
